from tombo import tombo_helper as th
from tombo._plot_commands import plot_single_read, test_r_imports

test_r_imports()

# plot single read
read_fn = 'path/to/read.fast5'
plot_single_read(fast5_fn=read_fn)


# plot first 10k raw obs from several reads
num_reads = 10
reads_index = th.TomboReads(['path/to/fast5/basedir/',])
for i, r_data in enumerate(reads_index.iter_reads()):
    plot_single_read(
        fast5_fn=r_data.fn, num_obs=10000,
        png_fn='single_read_raw_signal.num_' + str(i) + '.png')
    if i > num_reads: break


# plot best reads
best_reads = sorted(
    (r_data.sig_match_score, r_data)
    for r_data in reads_index.iter_reads())[:num_reads]
for i, (r_score, r_data) in enumerate(best_reads):
    plot_single_read(
        fast5_fn=r_data.fn, num_obs=10000,
        png_fn='best_reads.num_' + str(i) + '.png')


# plot worst scoring reads
reads_index = th.TomboReads(['path/to/fast5/basedir/',], remove_filtered=False)
worst_reads = sorted(
    (r_data.sig_match_score, r_data)
    for r_data in reads_index.iter_reads())[::-1][:num_reads]
for i, (r_score, r_data) in enumerate(worst_reads):
    plot_single_read(
        fast5_fn=r_data.fn, num_obs=10000,
        png_fn='worst_reads.num_' + str(i) + '.png')



# _c_dynamic_programming.pyx
# cython: profile=True

cimport cython

import numpy as np
cimport numpy as np

DTYPE = np.float64
ctypedef np.float64_t DTYPE_t

DTYPE_INT = np.int64
ctypedef np.int64_t DTYPE_INT_t

from libcpp cimport bool

def c_base_z_scores(
        np.ndarray[DTYPE_t] b_sig not None, DTYPE_t ref_mean, DTYPE_t ref_sd,
        bool do_winsorize_z=False, DTYPE_t max_half_z_score=10.0):
    cdef DTYPE_INT_t n_sig = b_sig.shape[0]
    b_z_scores = np.empty(n_sig, dtype=DTYPE)
    cdef DTYPE_t b_pos_z_score
    cdef DTYPE_INT_t idx
    for idx in range(n_sig):
        b_pos_z_score = (b_sig[idx] - ref_mean) / ref_sd
        if b_pos_z_score > 0:
            # convert all z-scores to lower tail
            b_pos_z_score = -b_pos_z_score
        if do_winsorize_z and b_pos_z_score < -max_half_z_score:
            b_pos_z_score = -max_half_z_score
        b_z_scores[idx] = b_pos_z_score
    return b_z_scores

def c_reg_z_scores(
        np.ndarray[DTYPE_t] r_sig not None,
        np.ndarray[DTYPE_t] r_ref_means not None,
        np.ndarray[DTYPE_t] r_ref_sds not None,
        np.ndarray[DTYPE_INT_t] r_b_starts not None,
        DTYPE_INT_t reg_start, DTYPE_INT_t reg_end,
        DTYPE_INT_t max_base_shift, DTYPE_INT_t min_obs_per_base,
        max_half_z_score=None):
    # check whether max_half_z_score is valid and set bool flag accordingly
    cdef DTYPE_t np_max_half_z_score
    cdef bool do_winsorize_z = False
    if max_half_z_score is not None:
        do_winsorize_z = True
        np_max_half_z_score = max_half_z_score

    cdef DTYPE_INT_t base_i, b_sig_start, b_sig_end, prev_sig_start, \
        prev_sig_end, idx
    cdef DTYPE_INT_t reg_len = reg_end - reg_start
    cdef np.ndarray[DTYPE_INT_t] sig_starts = np.empty(reg_len, dtype=DTYPE_INT)
    prev_start_set = False
    cdef np.ndarray[DTYPE_INT_t] base_range = np.arange(
        reg_start, reg_end, dtype=DTYPE_INT)
    for idx in range(reg_len):
        base_i = base_range[idx]
        b_sig_start = r_b_starts[max(reg_start, base_i - max_base_shift)]
        # clip observations from the beginning of a base if there is no
        # possible traceback path through that location
        if (prev_start_set and
            b_sig_start < prev_sig_start + min_obs_per_base):
            b_sig_start = prev_sig_start + min_obs_per_base
        prev_start_set = True
        sig_starts[idx] = b_sig_start
        prev_sig_start = b_sig_start

    cdef np.ndarray[DTYPE_INT_t] sig_ends = np.empty(reg_len, dtype=DTYPE_INT)
    prev_end_set = False
    # clip positions from the end of each base
    for idx in range(reg_len):
        base_i = base_range[reg_len - idx - 1]
        b_sig_end = r_b_starts[min(reg_end, base_i + max_base_shift + 1)]
        # clip observations from the end of a base if there is no
        # possible traceback path through that location
        if (prev_end_set and
            b_sig_end > prev_sig_end - min_obs_per_base):
            b_sig_end = prev_sig_end - min_obs_per_base
        prev_end_set = True
        sig_ends[reg_len - idx - 1] = b_sig_end
        prev_sig_end = b_sig_end

    reg_scores = []
    for idx in range(reg_len):
        base_i = base_range[idx]
        b_sig_start = sig_starts[idx]
        b_sig_end = sig_ends[idx]
        # z-score computation is far more efficient than p-values and
        # produces *very* similar results
        reg_scores.append((
            c_base_z_scores(r_sig[b_sig_start:b_sig_end],
                            r_ref_means[base_i], r_ref_sds[base_i],
                            do_winsorize_z, np_max_half_z_score), (
                                b_sig_start - r_b_starts[reg_start],
                                b_sig_end - r_b_starts[reg_start])))

    return reg_scores

def c_base_forward_pass(
        np.ndarray[DTYPE_t] b_data not None,
        DTYPE_INT_t b_start, DTYPE_INT_t b_end,
        np.ndarray[DTYPE_t] prev_b_data not None,
        DTYPE_INT_t prev_b_start, DTYPE_INT_t prev_b_end,
        np.ndarray[DTYPE_t] prev_b_fwd_data not None,
        np.ndarray[DTYPE_INT_t] prev_b_last_diag not None,
        DTYPE_INT_t min_obs_per_base):
    cdef DTYPE_INT_t b_len = b_end - b_start
    # forward pass cumulative z-scores for this base
    cdef np.ndarray[DTYPE_t] b_fwd_data = np.empty(b_len, dtype=DTYPE)
    # store last diagonal move to pass on to next base
    cdef np.ndarray[DTYPE_INT_t] b_last_diag = np.empty(b_len, dtype=DTYPE_INT)
    # use cumsum as it is much more efficient than sums
    cdef np.ndarray[DTYPE_t] prev_b_data_cumsum = np.cumsum(prev_b_data)
    cdef DTYPE_INT_t pos, last_valid_diag_lag, pos_diag_val
    cdef DTYPE_t diag_score, stay_base_score, pos_score, fwd_value

    # add the diagonally below position value for the first possible
    # position in each base
    fwd_value = b_data[0] + prev_b_fwd_data[b_start - prev_b_start - 1]
    b_fwd_data[0] = fwd_value
    b_last_diag[0] = 1

    # some bases end at the same position (could change this by trimming earlier)
    for pos in range(b_start + 1, prev_b_end + 1):
        last_valid_diag_lag = 1
        while (prev_b_last_diag[pos - prev_b_start - last_valid_diag_lag] +
               last_valid_diag_lag <= min_obs_per_base):
            last_valid_diag_lag += 1

        # about 50% of the model_resqiuiggle algorithm is spent on this
        # part of this loop
        # sum of last valid fwd position and original z-scores through to the
        # position before current pos
        diag_score = prev_b_fwd_data[pos - prev_b_start - last_valid_diag_lag]
        if last_valid_diag_lag > 1:
            diag_score += prev_b_data_cumsum[pos - prev_b_start - 1] - \
                          prev_b_data_cumsum[pos - prev_b_start -
                                             last_valid_diag_lag]
        stay_base_score = b_fwd_data[pos - b_start - 1]
        if diag_score > stay_base_score:
            pos_score, pos_diag_val = diag_score, 1
        else:
            # stayed in this base, so add one to the last stayed in base count
            pos_score, pos_diag_val = (
                stay_base_score, b_last_diag[pos - b_start - 1] + 1)
        fwd_value = b_data[pos - b_start] + pos_score
        b_fwd_data[pos - b_start] = fwd_value
        b_last_diag[pos - b_start] = pos_diag_val

    cdef DTYPE_INT_t idx, curr_last_diag, reg_left_len
    if b_end > prev_b_end + 1:
        # perform C cumsum until the end of the base
        # note no possible allowed diagonal moves here
        fwd_value = b_fwd_data[prev_b_end - b_start]
        curr_last_diag = b_last_diag[prev_b_end - b_start]
        reg_left_len = b_end - prev_b_end - 1
        for idx in range(reg_left_len):
            fwd_value += b_data[idx + prev_b_end - b_start + 1]
            curr_last_diag += 1
            b_fwd_data[idx + prev_b_end - b_start + 1] = fwd_value
            b_last_diag[idx + prev_b_end - b_start + 1] = curr_last_diag

    return b_fwd_data, b_last_diag

def c_base_traceback(
        np.ndarray[DTYPE_t] curr_b_data not None, DTYPE_INT_t curr_start,
        np.ndarray[DTYPE_t] next_b_data not None,
        DTYPE_INT_t next_start, DTYPE_INT_t next_end,
        DTYPE_INT_t sig_start, DTYPE_INT_t min_obs_per_base):
    cdef DTYPE_INT_t curr_base_sig = 1
    cdef DTYPE_INT_t sig_pos
    for sig_pos in range(sig_start, -1, -1):
        curr_base_sig += 1
        # if there is not enough signal in the current base or the next base
        # hasn't started yet continue on to the next signal position
        if curr_base_sig <= min_obs_per_base or sig_pos - 1 >= next_end:
            continue
        # if the current base has ended or the next base has a better score
        if (sig_pos <= curr_start or
            next_b_data[sig_pos-next_start-1] >
            curr_b_data[sig_pos-curr_start-1]):
            return sig_pos

@cython.wraparound(False)
@cython.boundscheck(False)
cdef DTYPE_INT_t c_argmax(np.ndarray[DTYPE_t] vals):
    cdef DTYPE_t val
    cdef DTYPE_t max_val = vals[0]
    cdef DTYPE_INT_t pos
    cdef DTYPE_INT_t max_pos = 0

    for pos in range(1, vals.shape[0]):
        val = vals[pos]
        if val > max_val:
            max_val = val
            max_pos = pos
    return max_pos

# Eventless re-squiggle dynamic programming algorithm
@cython.wraparound(False)
@cython.boundscheck(False)
cdef void c_process_band(
        np.ndarray[DTYPE_t, ndim=2] fwd_pass,
        np.ndarray[DTYPE_INT_t, ndim=2] fwd_pass_tb,
        np.ndarray[DTYPE_t] shifted_z_scores,
        DTYPE_t stay_pen, DTYPE_t skip_pen,
        DTYPE_INT_t bandwidth, DTYPE_INT_t band_starts_diff,
        DTYPE_INT_t seq_pos):

    cdef DTYPE_INT_t band_pos, max_from, prev_b_pos
    cdef DTYPE_t max_score, diag_score, skip_score, pos_z_score

    for band_pos in range(1, bandwidth):
        pos_z_score = shifted_z_scores[band_pos]
        prev_b_pos = band_pos + band_starts_diff

        # first set to stay state
        max_score = fwd_pass[seq_pos+1, band_pos-1] - stay_pen + pos_z_score
        max_from = 0
        # then check diagonal score
        if prev_b_pos - 1 < bandwidth:
            diag_score = fwd_pass[seq_pos, prev_b_pos-1] + pos_z_score
            if diag_score > max_score:
                max_score = diag_score
                max_from = 2
            # finally check skip score (note nested check to save some ops)
            if prev_b_pos < bandwidth:
                skip_score = fwd_pass[seq_pos, prev_b_pos] - skip_pen
                if skip_score > max_score:
                    max_score = skip_score
                    max_from = 1

        fwd_pass[seq_pos + 1, band_pos] = max_score
        fwd_pass_tb[seq_pos + 1, band_pos] = max_from

    return

@cython.wraparound(False)
@cython.boundscheck(False)
def c_banded_forward_pass(
        np.ndarray[DTYPE_t, ndim=2] shifted_z_scores not None,
        np.ndarray[DTYPE_INT_t, ndim=1] event_starts not None,
        DTYPE_t skip_pen, DTYPE_t stay_pen):
    cdef DTYPE_INT_t n_bases = shifted_z_scores.shape[0]
    cdef DTYPE_INT_t bandwidth = shifted_z_scores.shape[1]
    cdef np.ndarray[DTYPE_t, ndim=2] fwd_pass = np.empty((
        n_bases + 1, bandwidth))
    cdef np.ndarray[DTYPE_INT_t, ndim=2] fwd_pass_tb = np.empty(
        (n_bases + 1, bandwidth), dtype=DTYPE_INT)
    # zero starts let the read start anywhere along the beginning
    # (for finding the read start)
    cdef DTYPE_INT_t idx
    for idx in range(bandwidth):
        fwd_pass[0, idx] = 0.0

    cdef DTYPE_INT_t max_from, band_pos, seq_pos, prev_b_pos, band_starts_diff
    cdef DTYPE_t max_score, pos_z_score, skip_score, diag_score

    for seq_pos in range(n_bases):
        # set first band position to skip score if the bands have the same start
        if seq_pos == 0 or event_starts[seq_pos] == event_starts[seq_pos-1]:
            fwd_pass[seq_pos + 1, 0] = fwd_pass[seq_pos, 0] - skip_pen
            fwd_pass_tb[seq_pos + 1, 0] = 1
        # else use the match score
        else:
            fwd_pass[seq_pos + 1, 0] = (
                fwd_pass[seq_pos, event_starts[seq_pos] -
                         event_starts[seq_pos-1] - 1] +
                shifted_z_scores[seq_pos, 0])
            fwd_pass_tb[seq_pos + 1, 0] = 2

        band_starts_diff = (
            event_starts[seq_pos] - event_starts[seq_pos-1]
            if seq_pos > 0 else 0)
        c_process_band(
            fwd_pass, fwd_pass_tb, shifted_z_scores[seq_pos,:], stay_pen,
            skip_pen, bandwidth, band_starts_diff, seq_pos)

    return fwd_pass, fwd_pass_tb

def c_banded_traceback(
        np.ndarray[DTYPE_INT_t, ndim=2] fwd_pass_tb not None,
        np.ndarray[DTYPE_INT_t] event_starts not None, DTYPE_INT_t band_pos,
        DTYPE_INT_t band_boundary_thresh=-1):
    # first row in fwd pass is a pseudo-row and does not represent a base
    cdef DTYPE_INT_t n_bases = fwd_pass_tb.shape[0] - 1
    cdef DTYPE_INT_t bandwidth = fwd_pass_tb.shape[1]
    cdef np.ndarray[DTYPE_INT_t] seq_poss = np.empty(
        n_bases + 1, dtype=DTYPE_INT)
    cdef DTYPE_INT_t curr_event_pos = band_pos + event_starts[n_bases - 1]
    # last position is the end of the current looking window which is the
    # passed value
    seq_poss[n_bases] = curr_event_pos + 1
    cdef DTYPE_INT_t curr_seq_pos
    for curr_seq_pos in range(n_bases, 0, -1):
        band_pos = curr_event_pos - event_starts[curr_seq_pos-1]
        # 0 indicates stay in the current base
        while fwd_pass_tb[curr_seq_pos, band_pos] == 0:
            band_pos -= 1
        # if diagonal position
        if fwd_pass_tb[curr_seq_pos, band_pos] == 2:
            band_pos -= 1
        if (band_boundary_thresh >= 0 and
            min(band_pos, bandwidth - band_pos - 1) < band_boundary_thresh):
            raise NotImplementedError(
                'Read event to sequence alignment extends beyond bandwidth')
        curr_event_pos = event_starts[curr_seq_pos-1] + band_pos
        seq_poss[curr_seq_pos-1] = curr_event_pos + 1

    return seq_poss

@cython.wraparound(False)
@cython.boundscheck(False)
def c_adaptive_banded_forward_pass(
        np.ndarray[DTYPE_t, ndim=2] fwd_pass not None,
        np.ndarray[DTYPE_INT_t, ndim=2] fwd_pass_tb not None,
        np.ndarray[DTYPE_INT_t] event_starts not None,
        np.ndarray[DTYPE_t] event_means not None,
        np.ndarray[DTYPE_t] r_ref_means not None,
        np.ndarray[DTYPE_t] r_ref_sds not None,
        DTYPE_t z_shift, DTYPE_t skip_pen, DTYPE_t stay_pen,
        DTYPE_INT_t start_seq_pos, DTYPE_t mask_fill_z_score,
        bool do_winsorize_z, DTYPE_t max_half_z_score,
        bool return_z_scores=False):
    cdef DTYPE_INT_t n_bases = fwd_pass.shape[0] - 1
    cdef DTYPE_INT_t bandwidth = fwd_pass.shape[1]
    cdef DTYPE_INT_t half_bandwidth = bandwidth / 2
    cdef DTYPE_INT_t n_events = event_means.shape[0]

    # comment out when profiling
    #cdef DTYPE_INT_t band_pos, max_from, prev_b_pos
    #cdef DTYPE_t max_score, diag_score, skip_score, pos_z_score

    cdef DTYPE_INT_t event_pos, seq_pos, prev_band_start, curr_band_start
    cdef DTYPE_t pos_z_score, ref_mean, ref_sd

    cdef np.ndarray[DTYPE_t] shifted_z_scores = np.empty(bandwidth, dtype=DTYPE)
    cdef np.ndarray[DTYPE_t, ndim=2] all_shifted_z_scores
    if return_z_scores:
        all_shifted_z_scores = np.empty((n_bases - start_seq_pos, bandwidth),
                                        dtype=DTYPE)
    for seq_pos in range(start_seq_pos, n_bases):
        # determine adaptive location for this sequence position
        prev_band_start = event_starts[seq_pos - 1]
        curr_band_start = prev_band_start + c_argmax(fwd_pass[seq_pos]) \
                          - half_bandwidth + 1
        if curr_band_start < prev_band_start:
            curr_band_start = prev_band_start
        if curr_band_start >= n_events:
            # if this isn't within one of the last sequence position
            # the read is forced to skip to the end and will likely
            # not end in a favorable alignment
            if seq_pos < n_bases - 2:
                raise NotImplementedError(
                    'Adaptive signal to seqeunce alignment extended ' +
                    'beyond raw signal')
            curr_band_start = n_events - 1
        event_starts[seq_pos] = curr_band_start

        # get shifted z-scores for adaptive band
        ref_mean = r_ref_means[seq_pos]
        ref_sd = r_ref_sds[seq_pos]
        if curr_band_start + bandwidth <= n_events:
            for event_pos in range(curr_band_start,
                                   curr_band_start + bandwidth):
                pos_z_score = (event_means[event_pos] - ref_mean) / ref_sd
                if pos_z_score < 0:
                    pos_z_score = -pos_z_score
                if do_winsorize_z:
                    pos_z_score = min(pos_z_score, max_half_z_score)
                shifted_z_scores[
                    event_pos - curr_band_start] = z_shift - pos_z_score
        else:
            # if this band extends beyond events array pad with low score
            # note in the unlikely event that the traceback path goes through
            # this region it will be clipped later
            for event_pos in range(curr_band_start, n_events):
                pos_z_score = (event_means[event_pos] - ref_mean) / ref_sd
                if pos_z_score < 0:
                    pos_z_score = -pos_z_score
                if do_winsorize_z:
                    pos_z_score = min(pos_z_score, max_half_z_score)
                shifted_z_scores[
                    event_pos - curr_band_start] = z_shift - pos_z_score
            for event_pos in range(n_events - curr_band_start, bandwidth):
                shifted_z_scores[event_pos] = mask_fill_z_score
        if return_z_scores:
            all_shifted_z_scores[seq_pos - start_seq_pos,:] = shifted_z_scores

        # now perform dynamic programming fill for this seq position

        # set first band position to skip score if the bands have the same start
        if curr_band_start == prev_band_start:
            fwd_pass[seq_pos + 1, 0] = fwd_pass[seq_pos, 0] - skip_pen
            fwd_pass_tb[seq_pos + 1, 0] = 1
        # else use the match score
        else:
            fwd_pass[seq_pos + 1, 0] = fwd_pass[
                seq_pos, curr_band_start - prev_band_start - 1] + \
                shifted_z_scores[0]
            fwd_pass_tb[seq_pos + 1, 0] = 2

        # profiling shows that >60% of the time is spent here. Not
        # functionalized now due to function call overheads
        c_process_band(
            fwd_pass, fwd_pass_tb, shifted_z_scores, stay_pen, skip_pen,
            bandwidth, curr_band_start - prev_band_start, seq_pos)

    if return_z_scores:
        return all_shifted_z_scores

    return



from __future__ import division, unicode_literals, absolute_import

from builtins import int, range, dict, map, zip

import sys

# Future warning from cython in h5py
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
import h5py

import numpy as np

from operator import itemgetter

if sys.version_info[0] > 2:
    unicode = str

from . import tombo_helper as th


VERBOSE = False


##############################
###### Filter Functions ######
##############################

def clear_filters(fast5s_dir, corr_grp):
    """Clear filters applied to this directories index files
    """
    reads_index = th.TomboReads([fast5s_dir,], corr_grp, remove_filtered=False)

    th.status_message('Clearing all filters.')
    reads_index.replace_index(dict(
        (chrm_strand, [rd._replace(filtered=False) for rd in cs_reads_index])
        for chrm_strand, cs_reads_index in reads_index))

    reads_index.write_index_file()
    th.status_message('All filters successfully cleared!')

    return

def print_filter_mess(
        num_filt_reads, prev_unfilt_reads, total_reads, fast5s_dir, filter_text):
    if prev_unfilt_reads == 0:
        th.error_message_and_exit(
            'No unfiltered reads present in current Tombo index.')

    th.status_message(
        'Filtered {:d} reads ({:.1%} of previously filtered and '.format(
            num_filt_reads, float(num_filt_reads) / prev_unfilt_reads) +
        '{:.1%} of all valid reads)'.format(
            float(num_filt_reads) / total_reads) +
        ' reads due to ' + filter_text + ' filter from ' + fast5s_dir + '.')
    return

def filter_reads_for_stuck(fast5s_dir, corr_grp, obs_filter):
    """Filter reads based on some observation per base threshold criteria
    """
    def read_is_stuck(fast5_fn, s_grp):
        try:
            with h5py.File(fast5_fn, 'r') as fast5_data:
                base_lens = th.get_single_slot_read_centric(
                    fast5_data, 'length', s_grp)
                if base_lens is None: return True
                return any(np.percentile(base_lens, pctl) > thresh
                           for pctl, thresh in obs_filter)
        except:
            return True


    reads_index = th.TomboReads([fast5s_dir,], corr_grp, remove_filtered=False)

    th.status_message('Filtering stuck reads.')
    filt_reads_index = {}
    prev_unfilt_reads, num_filt_reads, total_reads = 0, 0, 0
    for chrm_strand, cs_reads in reads_index:
        cs_filt_reads = []
        for rd in cs_reads:
            total_reads += 1
            if not rd.filtered:
                prev_unfilt_reads += 1
                if read_is_stuck(rd.fn, rd.corr_group):
                    num_filt_reads += 1
                    rd = rd._replace(filtered=True)
            cs_filt_reads.append(rd)
        filt_reads_index[chrm_strand] = cs_filt_reads

    print_filter_mess(num_filt_reads, prev_unfilt_reads, total_reads,
                      fast5s_dir, 'observations per base')

    reads_index.replace_index(filt_reads_index)
    reads_index.write_index_file()

    return

def filter_reads_for_coverage(fast5s_dir, corr_grp, frac_to_filter):
    """Filter reads at higher coverage regions
    """
    reads_index = th.TomboReads([fast5s_dir,], corr_grp, remove_filtered=False)

    th.status_message('Filtering reads to obtain more uniform coverage.')
    filt_reads_index = dict((cs, []) for cs in reads_index.get_all_cs())
    unfilt_reads = []
    unfilt_reads_cov = []
    total_reads = 0
    for chrm_strand, cs_reads in reads_index:
        # TODO: perform coverage computation iteratively for larger fractions
        # of reads requested to avoid canceling out high coverage locations

        # compute coverage
        max_end = max(rd.end for rd in cs_reads)
        cs_coverage = np.zeros(max_end, dtype=np.int64)
        for rd in cs_reads:
            total_reads += 1
            if rd.filtered:
                # add previously filtered reads straight back to new index
                filt_reads_index[chrm_strand].append(rd)
            cs_coverage[rd.start:rd.end] += 1
        # assign coverage value to each read
        for rd in (rd for rd in cs_reads if not rd.filtered):
            # add approximate coverage from middle of read
            # faster than mean over the whole read
            unfilt_reads_cov.append(cs_coverage[
                rd.start + ((rd.end - rd.start) // 2)])
            unfilt_reads.append((chrm_strand, rd))

    prev_unfilt_reads = len(unfilt_reads)
    if prev_unfilt_reads == 0:
        th.error_message_and_exit(
            'No unfiltered reads present in current Tombo index.')
    num_filt_reads = int(frac_to_filter * prev_unfilt_reads)
    print_filter_mess(num_filt_reads, prev_unfilt_reads, total_reads,
                      fast5s_dir, 'even coverage')

    # create probabilities array with coverage values normalized to sum to 1
    unfilt_reads_cov = np.array(unfilt_reads_cov, dtype=np.float)
    unfilt_reads_p = unfilt_reads_cov / unfilt_reads_cov.sum()
    # randomly chose reads to filter
    filt_indices = set(np.random.choice(
        prev_unfilt_reads, size=num_filt_reads, replace=False, p=unfilt_reads_p))
    for i, (chrm_strand, rd) in enumerate(unfilt_reads):
        if i in filt_indices:
            rd = rd._replace(filtered=True)
        filt_reads_index[chrm_strand].append(rd)

    reads_index.replace_index(filt_reads_index)
    reads_index.write_index_file()

    return

def filter_reads_for_qscore(fast5s_dir, bc_grp, corr_grp, q_score_thresh):
    """Filter reads based on mean q-score
    """
    def read_fails_q_score(fast5_fn, s_grp):
        try:
            with h5py.File(fast5_fn, 'r') as fast5_data:
                r_q_scores = fast5_data['/Analyses/' + bc_grp + '/' + s_grp +
                                        '/Fastq'][:].decode().split('\n')[3]
                return th.get_mean_q_score(r_q_scores) < q_score_thresh
        except:
            return True


    reads_index = th.TomboReads([fast5s_dir,], corr_grp, remove_filtered=False)

    th.status_message('Filtering reads below a mean q-score cutoff.')
    filt_reads_index = dict((cs, []) for cs in reads_index.get_all_cs())
    num_filt_reads, prev_unfilt_reads, total_reads = 0, 0, 0
    for chrm_strand, cs_reads in reads_index:
        for rd in cs_reads:
            total_reads += 1
            if rd.filtered:
                filt_reads_index[chrm_strand].append(rd)
                continue
            prev_unfilt_reads += 1
            if rd.mean_q_score is None:
                filter_read = read_fails_q_score(
                    rd.fn, rd.corr_group.split('/')[-1])
            else:
                filter_read = rd.mean_q_score < q_score_thresh
            if filter_read:
                num_filt_reads += 1
                rd = rd._replace(filtered=True)
            filt_reads_index[chrm_strand].append(rd)

    print_filter_mess(num_filt_reads, prev_unfilt_reads, total_reads,
                      fast5s_dir, 'q-score')

    reads_index.replace_index(filt_reads_index)
    reads_index.write_index_file()

    return

def filter_reads_for_signal_matching(fast5s_dir, corr_grp, sig_match_thresh):
    """Filter reads based on observed to expected signal matching score
    """
    def read_fails_matching_score(fast5_fn, corr_group):
        try:
            with h5py.File(fast5_fn, 'r') as fast5_data:
                return fast5_data['/Analyses/' + corr_group].attrs.get(
                    'signal_match_score') > sig_match_thresh
        except:
            return True


    reads_index = th.TomboReads([fast5s_dir,], corr_grp, remove_filtered=False)

    th.status_message('Filtering reads above a signal matching score threshold.')
    filt_reads_index = dict((cs, []) for cs in reads_index.get_all_cs())
    num_filt_reads, prev_unfilt_reads, total_reads = 0, 0, 0
    for chrm_strand, cs_reads in reads_index:
        for rd in cs_reads:
            total_reads += 1
            if rd.filtered:
                filt_reads_index[chrm_strand].append(rd)
                continue
            prev_unfilt_reads += 1
            if rd.sig_match_score is None:
                filter_read = read_fails_matching_score(rd.fn, rd.corr_group)
            else:
                filter_read = rd.sig_match_score > sig_match_thresh
            if filter_read:
                num_filt_reads += 1
                rd = rd._replace(filtered=True)
            filt_reads_index[chrm_strand].append(rd)

    print_filter_mess(num_filt_reads, prev_unfilt_reads, total_reads,
                      fast5s_dir, 'signal matching')

    reads_index.replace_index(filt_reads_index)
    reads_index.write_index_file()

    return

def filter_reads_for_genome_pos(
        fast5s_dir, corr_grp, include_regs, include_partial=False):
    """Filter reads to include or exclude genomic regions
    """
    def read_included(start, end, chrm_include_regs):
        if chrm_include_regs is None:
            return True
        if include_partial:
            # include all reads partially overlapping regions
            return any(not (start > i_end or end < i_start)
                       for i_start, i_end in chrm_include_regs)

        # test if read is completely contained within the interval
        return any((start >= i_start and end <= i_end)
                   for i_start, i_end in chrm_include_regs)


    reads_index = th.TomboReads([fast5s_dir,], corr_grp, remove_filtered=False)

    th.status_message('Filtering reads outside of the specified ' +
                      'genomic location.')
    filt_reads_index = dict((cs, []) for cs in reads_index.get_all_cs())
    num_filt_reads, prev_unfilt_reads, total_reads = 0, 0, 0
    for (chrm, strand), cs_reads in reads_index:
        do_filter_cs_reads = chrm not in include_regs
        for rd in cs_reads:
            total_reads += 1
            if rd.filtered:
                filt_reads_index[(chrm, strand)].append(rd)
                continue
            prev_unfilt_reads += 1
            if do_filter_cs_reads or not read_included(
                    rd.start, rd.end, include_regs[chrm]):
                num_filt_reads += 1
                rd = rd._replace(filtered=True)
            filt_reads_index[(chrm, strand)].append(rd)

    print_filter_mess(num_filt_reads, prev_unfilt_reads, total_reads,
                      fast5s_dir, 'genomic position')

    reads_index.replace_index(filt_reads_index)
    reads_index.write_index_file()

    return


###################################
###### Filter Main Functions ######
###################################

def _clear_filters_main(args):
    for fast5s_dir in args.fast5_basedirs:
        clear_filters(fast5s_dir, args.corrected_group)

    return

def _filter_stuck_main(args):
    obs_filter = th.parse_obs_filter(args.obs_per_base_filter)
    for fast5s_dir in args.fast5_basedirs:
        filter_reads_for_stuck(fast5s_dir, args.corrected_group, obs_filter)

    return

def _filter_coverage_main(args):
    if not 0 < args.percent_to_filter < 100:
        th.error_message_and_exit(
            '--percent-to-filter must be between 0 and 100.')

    for fast5s_dir in args.fast5_basedirs:
        filter_reads_for_coverage(
            fast5s_dir, args.corrected_group, args.percent_to_filter / 100.0)

    return

def _filter_q_score_main(args):
    if not 0 < args.q_score < 40:
        th.error_message_and_exit('--q-score must be between 0 and 40.')

    for fast5s_dir in args.fast5_basedirs:
        filter_reads_for_qscore(
            fast5s_dir, args.basecall_group, args.corrected_group, args.q_score)

    return

def _filter_signal_matching_main(args):
    if not 0 < args.signal_matching_score < 10:
        th.error_message_and_exit(
            '--signal-matching-score must be between 0 and 10.')

    for fast5s_dir in args.fast5_basedirs:
        filter_reads_for_signal_matching(
            fast5s_dir, args.corrected_group, args.signal_matching_score)

    return

def _filter_genome_pos_main(args):
    include_regs = th.parse_genome_regions(args.include_regions)

    for fast5s_dir in args.fast5_basedirs:
        filter_reads_for_genome_pos(
            fast5s_dir, args.corrected_group, include_regs,
            args.include_partial_overlap)

    return

def filter_main(args):
    global VERBOSE
    VERBOSE = not args.quiet
    th.VERBOSE = VERBOSE

    if args.action_command == 'clear_filters':
        _clear_filters_main(args)
    elif args.action_command == 'genome_locations':
        _filter_genome_pos_main(args)
    elif args.action_command == 'stuck':
        _filter_stuck_main(args)
    elif args.action_command == 'level_coverage':
        _filter_coverage_main(args)
    elif args.action_command == 'q_score':
        _filter_q_score_main(args)
    elif args.action_command == 'raw_signal_matching':
        _filter_signal_matching_main(args)
    else:
        th.error_message_and_exit('Invalid Tombo filter command.')

    return


if __name__ == '__main__':
    sys.stderr.write('This is a module. See commands with `tombo -h`')
    sys.exit(1)



from __future__ import division, unicode_literals, absolute_import

from builtins import int, range, dict, map, zip

import io
import os
import re
import sys
import queue

# Future warning from cython in h5py
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
import h5py

from tqdm import tqdm
from time import sleep
from itertools import islice
from multiprocessing import Process, Queue, Pipe

if sys.version_info[0] > 2:
    unicode = str

from . import tombo_helper as th


VERBOSE = False

_MAX_QUEUE_SIZE = 1000
_ITER_QUEUE_LIMIT = 1000
_PROC_UPDATE_INTERVAL = 100

_MAX_FASTQ_QUEUE_SIZE = 10000
_SEQ_SUMMARY_FN_FIELD = 'filename'
_SEQ_SUMMARY_ID_FIELD = 'read_id'

# warning messages for annotate with fastqs over multiple processes,
# requiring passing warning codes to only print warning once.
_WARN_ID_VAL = 'ids'
_WARN_IO_VAL = 'io'
_WARN_MISMATCH_VAL = 'mismatch'
_WARN_OVRWRT_VAL = 'overwrite'
_WARN_UNIQ_VAL = 'uniq'
_WARN_CODES = (_WARN_ID_VAL, _WARN_IO_VAL, _WARN_MISMATCH_VAL, _WARN_OVRWRT_VAL)
_WARN_CODES_PREP = (_WARN_OVRWRT_VAL, _WARN_UNIQ_VAL)
_WARN_PREFIX = '****** WARNING ****** '


##########################
###### Annotate Raw ######
##########################

def _prep_fast5_for_fastq(fast5_data, bc_grp_name, bc_subgrp_name, overwrite):
    read_id = th.get_raw_read_slot(fast5_data).attrs.get('read_id')
    try:
        read_id = read_id.decode()
    except (AttributeError, TypeError):
        pass
    if read_id is None:
        return

    # if Analyses group doesn't exist yet, create it
    try:
        analyses_grp = fast5_data['/Analyses']
    except:
        analyses_grp = fast5_data.create_group('Analyses')

    # create Fastq slot, unless value exists and --overwrite is not set
    try:
        bc_grp = analyses_grp[bc_grp_name]
        bc_subgrp = analyses_grp[bc_subgrp_name]
    except:
        try:
            bc_grp = analyses_grp.create_group(bc_grp_name)
            bc_subgrp = bc_grp.create_group(bc_subgrp_name)
        except:
            if overwrite:
                del analyses_grp[bc_grp_name]
                bc_grp = analyses_grp.create_group(bc_grp_name)
                bc_subgrp = bc_grp.create_group(bc_subgrp_name)
            else:
                raise th.TomboError(
                    bc_grp_name + ' exists and --overwrite is not set.')

    return read_id

def _annotate_with_fastqs_worker(
        fastq_rec_q, fast5s_read_ids, fastq_slot, fq_slot_prepped,
        prog_q, warn_q, bc_grp_name, bc_subgrp_name, overwrite):
    been_warned = dict((warn_code, False) for warn_code in _WARN_CODES)
    num_recs_proc = 0
    while True:
        fastq_rec = fastq_rec_q.get()
        if fastq_rec is None:
            break

        # extract read_id from fastq (which should be the first text after
        # the "@" record delimiter up to the first white space or underscore
        read_id = fastq_rec[0].split()[0].split('_')[0][1:]
        if read_id not in fast5s_read_ids:
            if not been_warned[_WARN_ID_VAL]:
                been_warned[_WARN_ID_VAL] = True
                warn_q.put(_WARN_ID_VAL)
            continue

        try:
            with h5py.File(fast5s_read_ids[read_id], 'r+') as fast5_data:
                if not fq_slot_prepped:
                    try:
                        file_parsed_id = _prep_fast5_for_fastq(
                            fast5_data, bc_grp_name, bc_subgrp_name, overwrite)
                    except th.TomboError:
                        if not been_warned[_WARN_OVRWRT_VAL]:
                            been_warned[_WARN_OVRWRT_VAL] = True
                            warn_q.put(_WARN_OVRWRT_VAL)
                        continue
                    if read_id != file_parsed_id:
                        if not been_warned[_WARN_MISMATCH_VAL]:
                            been_warned[_WARN_MISMATCH_VAL] = True
                            warn_q.put(_WARN_MISMATCH_VAL)
                        continue
                bc_slot = fast5_data[fastq_slot]
                # add sequence to fastq slot
                bc_slot.create_dataset(
                    'Fastq', data=''.join(fastq_rec),
                    dtype=h5py.special_dtype(vlen=unicode))

                # progress q update
                num_recs_proc += 1
                if num_recs_proc % _PROC_UPDATE_INTERVAL == 0:
                    prog_q.put(_PROC_UPDATE_INTERVAL)
        except:
            if not been_warned[_WARN_IO_VAL]:
                been_warned[_WARN_IO_VAL] = True
                warn_q.put(_WARN_IO_VAL)
            continue

    # add last number of records reported from this process
    prog_q.put(num_recs_proc % _PROC_UPDATE_INTERVAL)

    return

def _feed_seq_records_worker(fastq_fns, fastq_rec_q, num_processes):
    for fastq_fn in fastq_fns:
        n_recs = 0
        with io.open(fastq_fn) as fastq_fp:
            while True:
                fastq_rec = list(islice(fastq_fp, 4))
                # if record contains fewer than 4 lines this indicates the
                # EOF, so move to next file
                if len(fastq_rec) != 4: break
                # if sequence identifier line does not start with "@" or quality
                # score line does not start with a "+" the file may be
                # corrupted, so don't process any more records
                if (re.match('@', fastq_rec[0]) is None or
                    re.match('\+', fastq_rec[2]) is None):
                    # TODO maybe send this as a warning code to avoid poorly
                    # formatted output
                    th.warning_message(
                        'Successfully parsed ' + unicode(n_recs) +
                        ' FASTQ records from ' + fastq_fn + ' before ' +
                        'encountering an invalid record. The rest of ' +
                        'this file will not be processed.')
                    break
                n_recs += 1
                fastq_rec_q.put(fastq_rec)

    # put none records to trigger annotation processes to exit
    for _ in range(num_processes):
        fastq_rec_q.put(None)

    return

def _get_ann_queues(prog_q, warn_q, num_read_ids, wp_conn):
    if VERBOSE: bar = tqdm(total=num_read_ids, smoothing=0)
    been_warned = dict((warn_code, False) for warn_code in _WARN_CODES)

    def update_warn(warn_val):
        if warn_val == _WARN_ID_VAL:
            if VERBOSE and not been_warned[_WARN_ID_VAL]:
                bar.write(
                    _WARN_PREFIX + 'Some FASTQ records contain read ' +
                    'identifiers not found in any FAST5 files or ' +
                    'sequencing summary files.',
                    file=sys.stderr)
            been_warned[_WARN_ID_VAL] = True
        elif warn_val == _WARN_IO_VAL:
            if VERBOSE and not been_warned[_WARN_IO_VAL]:
                bar.write(
                    _WARN_PREFIX + 'Some read files could not be accessed.',
                    file=sys.stderr)
            been_warned[_WARN_IO_VAL] = True
        elif warn_val == _WARN_MISMATCH_VAL:
            if VERBOSE and not been_warned[_WARN_MISMATCH_VAL]:
                bar.write(
                    _WARN_PREFIX + 'Read ID(s) found in sequencing summary ' +
                    'and FAST5 file are discordant. Skipping such reads.',
                    file=sys.stderr)
            been_warned[_WARN_MISMATCH_VAL] = True
        elif warn_val == _WARN_OVRWRT_VAL:
            if VERBOSE and not been_warned[_WARN_OVRWRT_VAL]:
                bar.write(
                    _WARN_PREFIX + 'Basecalls exsit in specified slot for ' +
                    'some reads. Set --overwrite option to overwrite these ' +
                    'basecalls.', file=sys.stderr)
            been_warned[_WARN_OVRWRT_VAL] = True
        else:
            if VERBOSE:
                bar.write('{}Invalid warning code encountered: {}'.format(
                    _WARN_PREFIX, warn_val), file=sys.stderr)

        return


    total_added_seqs = 0
    while True:
        try:
            iter_added = prog_q.get(block=False)
            total_added_seqs += iter_added
            if VERBOSE: bar.update(iter_added)
        except queue.Empty:
            try:
                warn_val = warn_q.get(block=False)
                update_warn(warn_val)
            except queue.Empty:
                sleep(0.1)
                # check if main thread has finished with all fastq records
                if wp_conn.poll():
                    break

    # collect all remaining warn and progress values
    while not prog_q.empty():
        iter_added = prog_q.get(block=False)
        total_added_seqs += iter_added
        if VERBOSE: bar.update(iter_added)
    while not warn_q.empty():
        warn_val = warn_q.get(block=False)
        update_warn(warn_val)

    if VERBOSE:
        bar.close()
        th.status_message('Added sequences to a total of ' +
                          str(total_added_seqs) + ' reads.')
        if total_added_seqs < num_read_ids:
            th.warning_message(
                'Not all read ids from FAST5s or sequencing summary files ' +
                'were found in FASTQs.\n\t\tThis can result from reads that ' +
                'failed basecalling or if full sets of FAST5s/sequence ' +
                'summaries are not processed with full sets of FASTQs.')

    return

def _annotate_with_fastqs(
        fastq_fns, fast5s_read_ids, fastq_slot, fq_slot_prepped, num_processes,
        bc_grp_name, bc_subgrp_name, overwrite):
    if VERBOSE: th.status_message('Annotating FAST5s with sequence from FASTQs.')
    fastq_rec_q = Queue(maxsize=_MAX_FASTQ_QUEUE_SIZE)
    prog_q = Queue()
    warn_q = Queue()

    # open a single process to read fastq files and feed the fastq record queue
    fq_feed_p = Process(target=_feed_seq_records_worker,
                        args=(fastq_fns, fastq_rec_q, num_processes))
    fq_feed_p.daemon = True
    fq_feed_p.start()

    # open fast5 annotation processes
    ann_args = (fastq_rec_q, fast5s_read_ids, fastq_slot, fq_slot_prepped,
                prog_q, warn_q, bc_grp_name, bc_subgrp_name, overwrite)
    ann_ps = []
    for p_id in range(num_processes):
        ann_p = Process(target=_annotate_with_fastqs_worker, args=ann_args)
        ann_p.daemon = True
        ann_p.start()
        ann_ps.append(ann_p)

    main_wp_conn, wp_conn = Pipe()
    warn_prog_p = Process(target=_get_ann_queues,
                          args=(prog_q, warn_q, len(fast5s_read_ids), wp_conn))
    warn_prog_p.daemon = True
    warn_prog_p.start()

    fq_feed_p.join()
    for ann_p in ann_ps:
        ann_p.join()
    # send signal to warn/progress queue that all other processes are complete
    main_wp_conn.send(True)
    warn_prog_p.join()

    return


##########################
#### Extract read_ids ####
##########################

def _get_prep_queue(read_ids_q, prog_q, warn_q, gp_conn, num_fast5s):
    """Process all records from all fast5 prep queues
    """
    ovrwrt_mess = (
        _WARN_PREFIX + 'Basecalls exsit in specified slot for some ' +
        'reads. Set --overwrite option to overwrite these basecalls.')
    fast5s_read_ids = {}
    # Warn non-unique read_ids in directory
    been_warned = dict((warn_code, False) for warn_code in _WARN_CODES_PREP)
    if VERBOSE: bar = tqdm(total=num_fast5s, smoothing=0)

    while True:
        try:
            read_id, fast5_fn = read_ids_q.get(block=False)
            if read_id in fast5s_read_ids:
                if VERBOSE and not been_warned[_WARN_UNIQ_VAL]:
                    bar.write(
                        _WARN_PREFIX + 'Multiple FAST5 files contain the ' +
                        'same read ID. Ensure that FAST5 files are from a ' +
                        'single run.', file=sys.stderr)
                    been_warned[_WARN_UNIQ_VAL] = True
                continue
            fast5s_read_ids[read_id] = fast5_fn
        except queue.Empty:
            try:
                warn_val = warn_q.get(block=False)
                if warn_val == _WARN_OVRWRT_VAL:
                    if VERBOSE and not been_warned[_WARN_OVRWRT_VAL]:
                        bar.write(ovrwrt_mess, file=sys.stderr)
                    been_warned[_WARN_OVRWRT_VAL] = True
                else:
                    bar.write('{}Invalid warning code encountered: {}'.format(
                        _WARN_PREFIX, warn_val), file=sys.stderr)
            except queue.Empty:
                try:
                    if VERBOSE: bar.update(prog_q.get(block=False))
                except queue.Empty:
                    sleep(0.1)
                    # check if main thread has finished with all FAST5s
                    if gp_conn.poll():
                        break

    while not read_ids_q.empty():
        read_id, fast5_fn = read_ids_q.get(block=False)
        fast5s_read_ids[read_id] = fast5_fn
    while not warn_q.empty():
        warn_val = warn_q.get(block=False)
        if warn_val == _WARN_OVRWRT_VAL:
            if VERBOSE and not been_warned[_WARN_OVRWRT_VAL]:
                bar.write(ovrwrt_mess, file=sys.stderr)
                been_warned[_WARN_OVRWRT_VAL] = True
        else:
            bar.write('{}Invalid warning code encountered: {}'.format(
                _WARN_PREFIX, warn_val), file=sys.stderr)
    while not prog_q.empty():
        if VERBOSE: bar.update(prog_q.get(block=False))

    if VERBOSE: bar.close()
    gp_conn.send(fast5s_read_ids)

    return

def _prep_fastq_slot_worker(
        fast5_q, bc_grp, bc_subgrp, overwrite, read_ids_q, prog_q, warn_q):
    num_files_proc = 0
    been_warned_overwrite = False
    while True:
        try:
            fast5_fn = fast5_q.get(block=False)
        except queue.Empty:
            sleep(0.1)
            continue

        if fast5_fn is None:
            break

        num_files_proc += 1
        if num_files_proc % _PROC_UPDATE_INTERVAL == 0:
            prog_q.put(_PROC_UPDATE_INTERVAL)

        try:
            with h5py.File(fast5_fn) as fast5_data:
                try:
                    read_id = _prep_fast5_for_fastq(
                        fast5_data, bc_grp, bc_subgrp, overwrite)
                except th.TomboError:
                    # avoid the warn queue getting too large by sending overwite
                    # warnings for each read from each thread
                    if not been_warned_overwrite:
                        been_warned_overwrite = True
                        warn_q.put(_WARN_OVRWRT_VAL)
                    continue
        except:
            continue
        if read_id is None:
            continue

        read_ids_q.put((read_id, fast5_fn))

    prog_q.put(num_files_proc % _PROC_UPDATE_INTERVAL)

    return

def _fill_files_queue(fast5_q, fast5_fns, num_ps):
    for fast5_fn in fast5_fns:
        fast5_q.put(fast5_fn)
    for _ in range(num_ps):
        fast5_q.put(None)

    return

def _get_read_ids_and_prep_fastq_slot(
        fast5s_dir, bc_grp, bc_subgrp, overwrite, num_processes):
    """Extract read id from /Raw group and prep fastq slots for annotation with
    associated FASTQ files.
    """
    if VERBOSE: th.status_message(
            'Preparing reads and extracting read identifiers.')
    fast5_q = Queue(maxsize=_MAX_QUEUE_SIZE)
    read_ids_q = Queue()
    prog_q = Queue()
    warn_q = Queue()

    fast5_fns = th.get_files_list(fast5s_dir)
    files_p = Process(target=_fill_files_queue,
                      args=(fast5_q, fast5_fns, num_processes))
    files_p.daemon = True
    files_p.start()

    prep_args = (fast5_q, bc_grp, bc_subgrp, overwrite, read_ids_q,
                 prog_q, warn_q)
    prep_ps = []
    for p_id in range(num_processes):
        prep_p = Process(target=_prep_fastq_slot_worker, args=prep_args)
        prep_p.daemon = True
        prep_p.start()
        prep_ps.append(prep_p)

    main_gp_conn, gp_conn = Pipe()
    get_prep_p = Process(
        target=_get_prep_queue,
        args=(read_ids_q, prog_q, warn_q, gp_conn, len(fast5_fns)))
    get_prep_p.daemon = True
    get_prep_p.start()

    # join all processes into the main thread
    files_p.join()
    for prep_p in prep_ps:
        prep_p.join()
    # send signal to get_prep queue that all other processes are complete
    main_gp_conn.send(True)
    fast5s_read_ids = main_gp_conn.recv()

    return fast5s_read_ids

def _parse_sequencing_summary_files(fast5s_dir, seq_summary_fns):
    if VERBOSE: th.status_message('Getting read filenames.')
    full_fast5_fns = {}
    # walk through directory structure searching for fast5 files
    for root, _, fns in os.walk(fast5s_dir):
        for fn in fns:
            if not fn.endswith('.fast5'): continue
            full_fast5_fns[fn] = os.path.join(root, fn)

    if VERBOSE: th.status_message('Parsing sequencing summary files.')
    fast5s_read_ids = {}
    been_warned = False
    for seq_summary_fn in seq_summary_fns:
        with open(seq_summary_fn) as fp:
            try:
                header_fields = fp.readline().split()
                fn_field = next(i for i, h_field in enumerate(header_fields)
                                if re.match(_SEQ_SUMMARY_FN_FIELD, h_field))
                id_field = next(i for i, h_field in enumerate(header_fields)
                                if re.match(_SEQ_SUMMARY_ID_FIELD, h_field))
            except:
                th.warning_message(
                    'Could not extract header information for sequencing ' +
                    'summary file: ' + seq_summary_fn)
                continue
            try:
                for line in fp:
                    rec_fields = line.split()
                    rec_short_fn = rec_fields[fn_field]
                    try:
                        rec_full_fn = full_fast5_fns[rec_short_fn]
                    except KeyError:
                        if not been_warned:
                            th.warning_message(
                                'Some FASTQ records from sequencing summaries ' +
                                'do not appear to have a matching file.')
                        been_warned = True
                        continue
                    # convert filename to full filename and link to read id
                    fast5s_read_ids[rec_fields[id_field]] = rec_full_fn
            except:
                th.warning_message(
                    'Error parsing records for sequencing ' +
                    'summary file: ' + seq_summary_fn)

    return fast5s_read_ids


##################################
###### Annotate FAST5s Main ######
##################################

def annotate_reads_with_fastq_main(args):
    global VERBOSE
    VERBOSE = not args.quiet
    th.VERBOSE = VERBOSE

    fast5s_basedir = (
        args.fast5_basedir if args.fast5_basedir.endswith('/') else
        args.fast5_basedir + '/')
    if args.sequencing_summary_filenames:
        fast5s_read_ids = _parse_sequencing_summary_files(
            fast5s_basedir, args.sequencing_summary_filenames)
        fq_slot_prepped = False
    else:
        fast5s_read_ids = _get_read_ids_and_prep_fastq_slot(
            fast5s_basedir, args.basecall_group, args.basecall_subgroup,
            args.overwrite, args.processes)
        fq_slot_prepped = True
    fastq_slot = '/'.join(('/Analyses', args.basecall_group,
                           args.basecall_subgroup))
    _annotate_with_fastqs(
        args.fastq_filenames, fast5s_read_ids, fastq_slot, fq_slot_prepped,
        args.processes, args.basecall_group, args.basecall_subgroup,
        args.overwrite)

    return


if __name__ == '__main__':
    sys.stderr.write('This is a module. See commands with `tombo -h`')
    sys.exit(1)



from __future__ import unicode_literals, absolute_import

from builtins import map

import sys
import argparse

if sys.version_info[0] > 2:
    unicode = str

from ._default_parameters import (
    SEG_PARAMS_TABLE, ALGN_PARAMS_TABLE, LLR_THRESH, SAMP_COMP_THRESH,
    DE_NOVO_THRESH, ALTERNATE_MODELS, MAX_SCALING_ITERS, ALT_EST_PCTL,
    COV_DAMP_COUNTS, SIG_MATCH_THRESH, FM_OFFSET_DEFAULT, OUTLIER_THRESH,
    DNA_SAMP_TYPE, RNA_SAMP_TYPE, MEAN_PRIOR_CONST, SD_PRIOR_CONST)

ALT_BASES = tuple(set(alt_name.split('_')[1] for alt_name in ALTERNATE_MODELS))


##################################
###### Positional arguments ######
##################################

basedir_opt=('fast5s_basedir', {
    'type':unicode,
    'help':'Directory containing fast5 files. All files ending in "fast5" ' +
    'found recursively within this base directory will be processed.'})
fasta_pos_opt=('reference', {
    'type':unicode, 'help':'Reference genome/transcriptome FASTA file ' +
    'or minimap2 index (with "map-ont" preset) for mapping.'})
fasta_event_opt=('reference_fasta', {
    'type':unicode, 'help':'Reference genome/transcriptome FASTA file ' +
    'for mapping.'})
# put re-squiggle positional arguments in one argument to allow printing
# hidden arguments help
rsqgl_pos_opt=('fast5s_and_reference', {
    'type':unicode, 'nargs':'*',
    'help':'Directory containing fast5 files and a genome/transcriptome ' +
    'reference. Directory will be searched recursively for files ending ' +
    'in ".fast5". Reference may be a FASTA file or minimap2 index file.'})


############################
###### Text arguments ######
############################

minimap2_opt=('--minimap2-executable', {
    'type':unicode, 'help':'Path to minimap2 executable.'})
minindx_opt=('--minimap2-index', {
    'type':unicode, 'help':'Path to minimap2 index (with map-ont preset) '
    'file corresponding to the [genome_fasta] provided.'})
bwamem_opt=('--bwa-mem-executable', {
    'type':unicode, 'help':'Path to bwa-mem executable.'})
graphmap_opt=('--graphmap-executable', {
    'type':unicode, 'help':'Path to graphmap executable.'})

poremod_opt=('--pore-model-filename', {
    'type':unicode,
    'help':'File containing kmer model parameters (level_mean ' +
    'and level_stdv) used in order to compute kmer-based corrected pA ' +
    'values. E.g. https://github.com/jts/nanopolish/blob/master/etc/' +
    'r9-models/template_median68pA.5mers.model'})
tbmod_opt=('--tombo-model-filename', {
    'type':unicode, 'help':'Tombo model filename. If no file is provided, ' +
    'the default DNA or RNA Tombo model will be used.'})
tbmod_w_opt=('--tombo-model-filename', {
    'type':unicode, 'help':'Filename to save Tombo model.'})
atbmod_opt=('--alternate-model-filename', {
    'type':unicode,
    'help':'Tombo model for alternative likelihood ratio significance testing.'})
hidden_tbmod_opt=('--tombo-model-filename', {
    'type':unicode, 'help':argparse.SUPPRESS})
hidden_atbmod_opt=('--alternate-model-filename', {
    'type':unicode, 'help':argparse.SUPPRESS})
hidden_atbmods_opt=('--alternate-model-filenames', {
    'type':unicode, 'nargs':'+', 'help':argparse.SUPPRESS})
altname_opt=('--alternate-model-name', {
    'type':unicode,
    'help':'A short name to associate with this alternate model (e.g. 5mC, ' +
    '6mA, etc.). This text will be included in output filenames when this ' +
    'model is used for testing.'})

failed_opt=('--failed-reads-filename', {
    'help':'Output failed read filenames with assoicated error. Default: ' +
    'Do not store failed reads.'})

sfast5dir_opt = ('--fast5-basedir', {
    'type':unicode, 'help':'Directory containing fast5 files.'})
fast5dir_opt = ('--fast5-basedirs', {
    'type':unicode, 'nargs':'+', 'help':'Directories containing fast5 files.'})
ctrlfast5dir_opt=('--control-fast5-basedirs', {
    'type':unicode, 'nargs':'+',
    'help':'Set of directories containing fast5 files for control reads, ' +
    'containing only canonical nucleotides.'})
altfast5dir_opt=('--alternate-fast5-basedirs', {
    'type':unicode, 'nargs':'+',
    'help':'Set of directories containing fast5 files for alternate set of ' +
    'reads.'})

corrgrp_opt=('--corrected-group', {
    'type':unicode, 'default':'RawGenomeCorrected_000',
    'help':'FAST5 group created by resquiggle command. Default: %(default)s'})
correvntgrp_opt=('--corrected-group', {
    'type':unicode, 'default':'RawGenomeCorrected_000',
    'help':'FAST5 group created by resquiggle command. Default: %(default)s'})
bcgrp_opt=('--basecall-group', {
    'type':unicode, 'default':'Basecall_1D_000',
    'help':'FAST5 group obtain original basecalls (under Analyses group). ' +
    'Default: %(default)s'})
bcsubgrps_opt=('--basecall-subgroups', {
    'type':unicode, 'default':['BaseCalled_template',], 'nargs':'+',
    'help':'FAST5 subgroup(s) (under /Analyses/[--basecall-group]/) ' +
    'containing basecalls and created within [--corrected-group] ' +
    'containing re-squiggle results. Default: %(default)s'})
bcsubgrp_opt=('--basecall-subgroup', {
    'type':unicode, 'default':'BaseCalled_template',
    'help':'FAST5 subgroup (under /Analyses/[--basecall-group]/) under which ' +
    'to store basecalls from FASTQs. Default: %(default)s'})

gnmloc_opt=('--genome-locations', {
    'type':unicode, 'nargs':'+',
    'help':'Genomic locations at which to plot signal. Format locations ' +
    'as "chrm:position[:strand] [chrm2:position2[:strand2] ...]" ' +
    '(strand not applicable for all applications)'})
incldreg_opt=('--include-regions', {
    'type':unicode, 'nargs':'+',
    'help':'Filter out reads not falling completely within include regions. ' +
    'Omit start and end coordinates to include an entire chromosome/sequence ' +
    'record. Format regions as "chrm[:start-end] [chrm2[:start2-end2] ...]".'})

fasta_opt=('--genome-fasta', {
    'type':unicode,
    'help':'FASTA file used to re-squiggle. For faster sequence access.'})
motif_opt=('--motif', {
    'type':unicode,
    'help':'Motif of interest at which to plot signal and statsitics. ' +
    'Supports IUPAC single letter codes (use T for RNA).'})

obsfilt_opt=('--obs-per-base-filter', {
    'type':unicode, 'nargs':'+', 'default':[],
    'help':'Filter reads based on observations per base percentile ' +
    'thresholds. Format thresholds as "percentile:thresh ' +
    '[pctl2:thresh2 ...]". For example to filter reads with 99th ' +
    'pctl > 200 obs/base or max > 5k obs/base use "99:200 100:5000".'})

fastqs_opt = ('--fastq-filenames', {
    'type':unicode, 'nargs':'+',
    'help':'FASTQ filenames containing basecalls to be added to ' +
    'raw FAST5 files.'})
seqsum_opt = ('--sequencing-summary-filenames', {
    'type':unicode, 'nargs':'+',
    'help':'Sequencing summary filenames produced by albacore. These can ' +
    'make annotation of raw FAST5 files with FASTQ sequence much faster.'})

brsrfn_opt=('--browser-file-basename', {
    'type':unicode, 'default':'tombo_results',
    'help':'Basename for output browser files. Two files (plus and minus ' +
    'strand) will be produced for each --file-types supplied. ' +
    'Filenames formatted as "[browser-file-basename].[file-type].' +
    '[sample|control]?.[plus|minus].[wig|bedgraph]". Default: %(default)s'})
pdf_opt=('--pdf-filename', {
    'type':unicode, 'help':'PDF filename to store plot(s). ' +
    'Default: %(default)s'})
statfn_opt=('--statistics-filename', {
    'type':unicode, 'help':'File to save/load genomic base anchored ' +
    'statistics.'})
statbsnm_opt=('--statistics-file-basename', {
    'type':unicode,
    'help':"File base name to save base by base statistics from testing. " +
    "Filenames will be [--statistics-file-basename]." +
    "[--alternate-bases]?.tombo.stats"})
rdata_opt=('--r-data-filename', {
    'type':unicode,
    'help':"Filename to save R data structure. Default: Don't save"})
seqs_opt=('--sequences-filename', {
    'type':unicode,
    'help':'File for sequences from selected regions. Sequences will be ' +
    'stored in FASTA format. Default: %(default)s.'})
densbn_opt=('--save-density-basename', {
    'type':unicode,
    'help':"Basename to save alternative model estimation density " +
    "estimation information. See scripts/debug_est_alt.R for info use " +
    "example. Default: Don't save."})
altden_opt=('--alternate-density-filename', {
    'type':unicode,
    'help':'File containing k-mer level kernel density estimates for the ' +
    'alternative sample saved using --save-density-basename.'})
ctrlden_opt=('--control-density-filename', {
    'type':unicode,
    'help':'File containing k-mer level kernel density estimates for the ' +
    'control sample saved using --save-density-basename.'})
prstatbn_opt=('--per-read-statistics-basename', {
    'type':unicode,
    'help':'Base for binary files containing per-read statistics from ' +
    'statistical testing. Filenames will be [--per-read-statistics-basename].' +
    '[--alternate-bases]?.tombo.per_read_stats'})
prstat_opt=('--per-read-statistics-filename', {
    'type':unicode,
    'help':'Binary file containing per-read statistics from ' +
    'statistical testing.'})
prstats_opt=('--per-read-statistics-filenames', {
    'type':unicode, 'nargs':'+',
    'help':'Binary files containing per-read statistics from ' +
    'statistical testing.'})
prctrlstats_opt=('--per-read-control-statistics-filenames', {
    'type':unicode, 'nargs':'+',
    'help':'Binary files containing per-read control statistics from ' +
    'statistical testing.'})

statfns_opt=('--statistics-filenames', {
    'type':unicode, 'nargs':'+',
    'help':"Files to load genomic base anchored statistics."})
ctrlstatfns_opt=('--control-statistics-filenames', {
    'type':unicode, 'nargs':'+', 'help':'Files to load genomic base anchored ' +
    'statistics from a control sample.'})
validlocs_opt=('--valid-locations-filename', {
    'type':unicode,
    'help':"Bed format file containing single base locations of valid sites. " +
    "Should contain 6 fields including strand. E.g. modified base locations."})
moddescs_opt=('--modified-locations', {
    'type':unicode, 'nargs':'+',
    'help':'Modification description and bed format files containing ' +
    'single base locations of ground truth modified sites. Bed files should ' +
    'contain 6 fields including strand. Format descriptions as ' +
    '"mod_name:locs.bed". Example: "CpG bisulfite":bisulfite_locs.bed'})
unmodlocs_opt=('--unmodified-locations', {
    'type':unicode, 'nargs':'+',
    'help':'Bed format files containing single base locations of ground ' +
    'truth unmodified sites. Bed files should contain 6 fields including ' +
    'strand.'})
motifdesc_opt=('--motif-description', {
    'type':unicode,
    'help':'Motif containing alternate-base. All positions with this motif ' +
    'should be modified (or filtered with [--valid-locations-filename]). ' +
    'Format descriptions as: "motif:mod_pos". mod_pos indicates the ' +
    'alternate-base within the motif (1-based index). Example: "CG:1" to ' +
    'train a CpG methylation model.'})
motifdescs_opt=('--motif-descriptions', {
    'type':unicode, 'nargs':'+',
    'help':'Ground truth, motif centered, modified base descriptions for ' +
    'computing ROC and PR curves. Each statistics file is associated with ' +
    'a set of motif descriptions. Format descriptions as: "motif:mod_pos:name' +
    '[::motif2:mod_pos2:name2...]". mod_pos indicates the alternate-base ' +
    'within the motif (1-based index). Example: CCWGG:2:"dcm 5mC"::GATC:2:' +
    '"dam 6mA" would assess the performance of a single Tombo statistics ' +
    'file for identification of E. coli dam and dcm methylation.'})
motifdescsimp_opt=('--motif-descriptions', {
    'type':unicode, 'nargs':'+',
    'help':'Ground truth, motif centered, modified base descriptions for ' +
    'output filtering. Format descriptions as: "motif:mod_pos:name". The ' +
    'mod_pos indicates the modified base within the motif (1-based index). ' +
    'Example: CCWGG:2:dcm_5mC GATC:2:dam_6mA would filter output for ' +
    'identification of E. coli dam and dcm methylation.'})


############################
###### Int arguments ######
############################

proc_opt=('--processes', {
    'type':int, 'help':'Number of processes. Default: %(default)d'})
thrpp_opt=('--threads-per-process', {
    'type':int, 'default':1,
    'help':'Number of file input/output and mapping threads per compute ' +
    'process [--processes]. This should likely be left at 1, but may ' +
    'improve performance on some systems. Default: %(default)d'})
hidthrpp_opt=('--threads-per-process', {
    'type':int, 'default':1, 'help':argparse.SUPPRESS})

alignproc_opt=('--align-processes', {
    'type':int, 'default':1,
    'help':'Number of processes to use for parsing and aligning ' +
    'original basecalls. Each process will independently load the ' +
    'genome into memory, so use caution with larger genomes ' +
    '(e.g. human). Default: %(default)d'})
alignthrds_opt=('--align-threads-per-process', {
    'type':int,
    'help':'Number of threads to use for aligner system call. ' +
    'Default: [--processes] / (2 * [--align-processes)]'})
rsqglproc_opt=('--resquiggle-processes', {
    'type':int,
    'help':'Number of processes to use for resquiggle algorithm. ' +
    'Default: [--processes] / 2'})
batchsize_opt=('--alignment-batch-size', {
    'default':1000, 'type':int,
    'help':'Number of reads included in each alignment call. ' +
    'Note: A new system mapping call is made for each batch ' +
    '(including loading of the genome), so it is advised to use ' +
    'larger values for larger genomes. Default: %(default)d'})
mpreg_opt=('--multiprocess-region-size', {
    'default':10000, 'type':int,
    'help':'Size of regions over which to multiprocesses statistic ' +
    'computation. For very deep samples a smaller value is recommmended ' +
    'in order to control memory consumption. Default: %(default)d'})
mstsgnf_opt=('--num-most-significant-stored', {
    'default':100000, 'type':int,
    'help':'Number of the most significant sites to store for faster access. ' +
    'If a longer list of most significant sites is required the list must be ' +
    're-computed from all batches. Very large values can increase RAM usage. ' +
    'Default: %(default)d'})

timeout_opt=('--timeout', {
    'type':int, 'help':'Timeout in seconds for processing a single read. ' +
    'Default: No timeout.'})
cpt_opt=('--cpts-limit', {
    'type':int, 'help':'Maximum number of changepoints to find within a ' +
    'single indel group. Default: No limit.'})

kmerthresh_opt=('--num-kmer-threshold', {
    'default':1, 'type':int,
    'help':'Observations of each k-mer required to include a read in ' +
    'read level averages. Default: %(default)d'})

covthresh_opt=('--coverage-threshold', {
    'type':int,
    'help':'Maximum mean coverage per region when estimating k-mer model ' +
    '(limits compute time for deep samples). Default: %(default)d'})
minkmer_opt=('--minimum-kmer-observations', {
    'type':int,
    'help':'Number of each k-mer observations required in order to produce ' +
    'a reference (genomic locations for standard reference and per-read ' +
    'for alternative reference). Default: %(default)d'})

numbases_opt=('--num-bases', {
    'type':int,
    'help':'Number of bases to plot/output. Default: %(default)d'})
numreads_opt=('--num-reads', {
    'type':int, 'help':'Number of reads to plot. Default: %(default)d'})
numreg_opt=('--num-regions', {
    'type':int, 'help':'Number of regions to plot. Default: %(default)d'})

slides_opt=('--slide-span', {
    'type':int, 'default':0,
    'help':'Number of bases offset over which to search when computing ' +
    'distances for signal cluster plotting. Default: 0 (exact position)'})
cntxt_opt=('--num-context', {
    'type':int, 'default':5,
    'help':'Number of context bases around motif. Default: %(default)d'})
numstat_opt=('--num-statistics', {
    'type':int, 'default':200,
    'help':'Number of motif centered regions to include in ' +
    'statistic distributions. Default: %(default)d'})
ovpltthresh_opt=('--overplot-threshold', {
    'type':int, 'default':50,
    'help':'Coverage level to trigger alternative plot type ' +
    'instead of raw signal. Default: %(default)d'})

fmo_opt=('--fishers-method-context', {
    'type':int, 'default':FM_OFFSET_DEFAULT,
    'help':'Number of context bases up and downstream over which to compute ' +
    "Fisher's method combined p-values. Note: Not applicable " +
    "for alternative model likelihood ratio tests. Default: %(default)d."})
minreads_opt=('--minimum-test-reads', {
    'type':int,
    'help':'Number of reads required at a position to perform significance ' +
    'testing or contribute to model estimation. Default: %(default)d'})

allspb_opt=('--statistics-per-block', {
    'type':int,
    'help':'Number of randomly selected per-read, per-base statistics to ' +
    'extract from each genomic block for plotting. Default: Include all stats'})
spb_opt=('--statistics-per-block', {
    'type':int,
    'help':'Number of randomly selected per-read, per-base statistics to ' +
    'extract from each genomic block for plotting. Default: %(default)d'})
tsl_opt=('--total-statistics-limit', {
    'type':int,
    'help':'Total per-read statistics to be extracted for plotting. ' +
    'Avoids memory overflow for large runs. Default: %(default)d'})

dynerr_opt=('--num-most-common-errors', {
    'type':int, 'default':0,
    'help':'Dynamically show this many most common errors so far through run. ' +
    'Default: 0; Just show progress'})
segpars_opt=('--segmentation-parameters', {
    'type':int, 'nargs':len(next(iter(SEG_PARAMS_TABLE.values()))),
    'help':'Specify parameters for segmentation 1) running neighboring ' +
    'windows width 2) minimum raw observations per genomic base, 3) raw ' +
    're-squiggle min obs per base 4) mean raw observations per event. ' +
    'Sample type defaults: ' +
    ' || '.join((bst + ' : ' + ' '.join(map(str, params)))
                for bst, params in SEG_PARAMS_TABLE.items())})
hidsegpars_opt=('--segmentation-parameters', {
    'type':int, 'nargs':len(next(iter(SEG_PARAMS_TABLE.values()))),
    'help':argparse.SUPPRESS})
segpars2_opt=('--segmentation-parameters', {
    'type':int, 'nargs':2,
    'help':'Specify the 2 parameters for segmentation 1) running neighboring ' +
    'windows width 2) minimum raw observations per genomic base. Sample type ' +
    'defaults:\n' +
    ' || '.join((bst + ' : ' + ' '.join(map(str, params[:2])))
                for bst, params in SEG_PARAMS_TABLE.items())})
msi_opt=('--max-scaling-iterations', {
    'type':int, 'default':MAX_SCALING_ITERS,
    'help':'Maximum re-squiggle iterations to perform. At each iteration ' +
    'the signal normalization parameters are re-fit. Higher values ' +
    'recommended for highly modified reads. Default: %(default)d'})
hidmsi_opt=('--max-scaling-iterations', {
    'type':int, 'default':MAX_SCALING_ITERS, 'help':argparse.SUPPRESS})
sigrng_opt=('--signal-length-range', {
    'type':int, 'nargs':2, 'help':'Allowed raw signal length range. ' +
    'Useful when reads with very long signal cause very slow processing or ' +
    'memory overflow. Reasonable maximum values may be around 500k. ' +
    'Default: No filtering'})
hidsigrng_opt=('--signal-length-range', {
    'type':int, 'nargs':2, 'help':argparse.SUPPRESS})
seqrng_opt=('--sequence-length-range', {
    'type':int, 'nargs':2, 'help':'Allowed mapped sequence length range. ' +
    'Useful when reads with very long signal cause very slow processing or ' +
    'memory overflow. Reasonable maximum values may be around 50k. ' +
    'Default: No filtering'})
hidseqrng_opt=('--sequence-length-range', {
    'type':int, 'nargs':2, 'help':argparse.SUPPRESS})


###############################
###### Boolean arguments ######
###############################

skpidx_opt=('--skip-index', {
    'default':False, 'action':'store_true',
    'help':'Skip creation of tombo index. This drastically slows downstream '+
    'tombo commands. Default stores tombo index named ".[--fast5-basedir].' +
    '[--corrected-group].tombo.index" to be loaded automatically for ' +
    'downstream commands.'})
hidskpidx_opt=('--skip-index', {
    'default':False, 'action':'store_true', 'help':argparse.SUPPRESS})
ovrwrt_opt=('--overwrite', {
    'default':False, 'action':'store_true',
    'help':'Overwrite previous corrected group in FAST5 files. Note: ' +
    'only effects --corrected-group or --new-corrected-group.'})
ignrlock_opt=('--ignore-read-locks', {
    'default':False, 'action':'store_true',
    'help':'Ignore Tombo locks, used to ensure that reads are only accessed ' +
    'from a single resquiggle processes avoiding potential file corruption.'})
hidignrlock_opt=('--ignore-read-locks', {
    'default':False, 'action':'store_true', 'help':argparse.SUPPRESS})
printadv_opt=('--print-advanced-arguments', {
    'default':False, 'action':'store_true',
    'help':'Print advanced re-squiggle arguments and exit.'})
printalt_opt=('--print-available-models', {
    'default':False, 'action':'store_true',
    'help':'Print available alternative models and exit.'})

estmean_opt=('--estimate-mean', {
    'default':False, 'action':'store_true',
    'help':"Use the mean instead of median for model level estimation. Note:" +
    " This can cause poor fits due to outliers"})
kmspec_opt=('--kmer-specific-sd', {
    'default':False, 'action':'store_true',
    'help':"Estimate standard deviation for each k-mers individually."})
incldsd_opt=('--include-event-stdev', {
    'default':False, 'action':'store_true',
    'help':'Include corrected event standard deviation in output FAST5 data.'})
hidincldsd_opt=('--include-event-stdev', {
    'default':False, 'action':'store_true', 'help':argparse.SUPPRESS})
fitscl_opt=('--fit-global-scale', {
    'default':False, 'action':'store_true',
    'help':'Fit a global scaling parameter for all reads. Otherwise fit ' +
    'the scaling parameter for each read. Global parameter estimated from ' +
    'a random subset of reads, which may produce more robust results for ' +
    'some samples.'})
hidfitscl_opt=('--fit-global-scale', {
    'default':False, 'action':'store_true', 'help':argparse.SUPPRESS})
sss_opt=('--skip-sequence-rescaling', {
    'default':False, 'action':'store_true',
    'help':'Skip sequence-based re-scaling. Otherwise, after re-squiggle, ' +
    'signal normalization parameters are re-fit (using Theil-Sen estimator).'})
hidsss_opt=('--skip-sequence-rescaling', {
    'default':False, 'action':'store_true', 'help':argparse.SUPPRESS})
stdllhr_opt=('--standard-log-likelihood-ratio', {
    'default':False, 'action':'store_true',
    'help':'Use a standard log likelihood ratio (LLR) statistic. Default ' +
    'is to use an outlier-robust LLR-like statistic. Detail in full ' +
    'online documentation.'})
prtovlp_opt=('--include-partial-overlap', {
    'default':False, 'action':'store_true',
    'help':'Include reads that partially overlap the specified region. ' +
    'Default: Only include reads completely contained in a specified region'})
storepval_opt=('--store-p-value', {
    'default':False, 'action':'store_true',
    'help':'Store p-value instead of effect-size statistic. Statistics are ' +
    'D-statistic (KS-test), deviation from even common language effect size ' +
    "(u-test), and Cohen's D (t-test)."})

readmean_opt=('--read-mean', {
    'default':False, 'action':'store_true',
    'help':'Plot k-mer means across whole reads as opposed to ' +
    'individual k-mer event levels.'})
boxc_opt=('--box-center', {
    'default':False, 'action':'store_true',
    'help':"Plot a box around the central base."})
deepcov_opt=('--deepest-coverage', {
    'default':False, 'action':'store_true',
    'help':'Plot the deepest coverage regions.'})
noplot_opt=('--dont-plot', {
    'default':False, 'action':'store_true',
    'help':"Don't plot result. Useful to produce only R data file."})
pstdmod_opt=('--plot-standard-model', {
    'default':False, 'action':'store_true',
    'help':"Add default standard model distribution to the plot."})
samponly_opt=('--sample-only-estimates', {
    'default':False, 'action':'store_true',
    'help':"Only use canonical sample to estimate expected signal level and " +
    "spread. Default: Use canonical model to improve estimtates (esp. for " +
    "low coverage regions) using baysian posterior estimates."})

quiet_opt=(('--quiet', '-q'), {
    'default':False, 'action':'store_true',
    'help':"Don't print status information."})


##############################
###### Float arguments ######
##############################

otlthresh_opt=('--outlier-threshold', {
    'type':float, 'default':OUTLIER_THRESH,
    'help':'Windosrize the signal at this number of scale values. ' +
    'Negative value disables outlier clipping. Default: %(default)f'})
hidotlthresh_opt=('--outlier-threshold', {
    'type':float, 'default':OUTLIER_THRESH, 'help':argparse.SUPPRESS})

snglrdthrsh_opt=('--single-read-threshold', {
    'type':float, 'nargs':'+',
    'help':(
        'P-value or log likelihood ratio threshold when computing ' +
        'fraction of significant reads at each genomic position. If two ' +
        'values are provided, statistics between these values are not ' +
        'considered.')})
dnthresh_opt=('--single-read-threshold', {
    'type':float, 'nargs':'+',
    'help':(
        'P-value threshold when computing fraction of significant reads at ' +
        'each genomic position. If two values are provided, statistics ' +
        'between these values are not considered. ' +
        'Default thresholds: ' +
        ', '.join(bst + ':' + (str(thresh[1]) if thresh[0] is None else
                               str(thresh[0]) + '-' + str(thresh[1])) + ' '
                  for bst, thresh in DE_NOVO_THRESH.items()))})
scompthresh_opt=('--single-read-threshold', {
    'type':float, 'nargs':'+',
    'help':(
        'P-value threshold when computing fraction of significant reads at ' +
        'each genomic position. If two values are provided, statistics ' +
        'between these values are not considered. ' +
        'Default thresholds: ' +
        ', '.join(bst + ':' + (str(thresh[1]) if thresh[0] is None else
                               str(thresh[0]) + '-' + str(thresh[1])) + ' '
                  for bst, thresh in SAMP_COMP_THRESH.items()))})
altthresh_opt=('--single-read-threshold', {
    'type':float, 'nargs':'+',
    'help':(
        'Log likelihood ratio threshold when computing fraction of ' +
        'significant reads at each genomic position. If two values ' +
        'are provided, statistics between these values are not considered. ' +
        'Default thresholds: ' +
        ', '.join(bst + ':' + (str(thresh[1]) if thresh[0] is None else
                               str(thresh[0]) + '-' + str(thresh[1])) + ' '
                  for bst, thresh in LLR_THRESH.items()))})

altfrac_opt=('--alt-fraction-percentile', {
    'default':ALT_EST_PCTL, 'type':float,
    'help':'When esitmating the alternative base incorporation rate, this ' +
    'percent of k-mers are assumed to have significantly shifted signal so ' +
    'the alternative distribution minimally overlaps the standard base ' +
    'distribution. Default: %(default)f'})
kernden_opt=('--kernel-density-bandwidth', {
    'default':0.05, 'type':float,
    'help':'Bandwidth applied when performing Gaussian kernal density ' +
    'esitmation on standard and alternative base signal distributions. ' +
    'Default: %(default)f'})
pctfilt_opt=('--percent-to-filter', {
    'type':float, 'default':10,
    'help':'Percentage of all reads to filter. Reads are randomly selected ' +
    'weighted according to the approximate coverage at the mapped genomic ' +
    'location. This can be useful in modeling and testing. ' +
    'Default: %(default)f'})
qscr_opt=('--q-score', {
    'type':float,
    'help':'Q-score threshold for filtering low quality reads. ' +
    'Default: %(default)f'})
sms_opt=('--signal-matching-score', {
    'type':float,
    'help':'Observed to expected signal matching score (higher score ' +
    'indicates poor matching). Sample type defaults: ' +
    ' || '.join(bst + ' : ' + str(params)
                for bst, params in SIG_MATCH_THRESH.items())})
fxdscl_opt=('--fixed-scale', {
    'type':float,
    'help':'Fixed scaling parameter to use for raw signal normalization.'})
hidfxdscl_opt=('--fixed-scale', {
    'type':float, 'help':argparse.SUPPRESS})
cvgdmp_opt=('--coverage-dampen-counts', {
    'type':float, 'nargs':2, 'default':COV_DAMP_COUNTS,
    'help':'Dampen fraction modified estimates for low coverage sites. Two ' +
    'parameters are unmodified and modified pseudo read counts. This is ' +
    'equivalent to a beta prior on the fraction estimate. Set to "0 0" to ' +
    'disable dampened fraction estimation. Default: %(default)s'})
prwht_opt=('--model-prior-weights', {
    'type':float, 'nargs':2, 'default':[MEAN_PRIOR_CONST, SD_PRIOR_CONST],
    'help':'Prior weights (one each for mean and spread) applied to ' +
    'canonical base model for estimating posterior model parameters for ' +
    'sample comparison. Default: %(default)s'})

sigapars_opt=('--signal-align-parameters', {
    'type':float, 'nargs':len(next(iter(ALGN_PARAMS_TABLE.values()))),
    'help':'Specify the parameters for signal to genome sequence alignment ' +
    'algorithm 1) match expected value 2) skip penalty 3) bandwidth 4) save ' +
    'bandwidth (if read fails with bandwith) 5) z-score winsorizing ' +
    'threshold 6) bandwidth boundary threshold 7) start bandwidth 8) start ' +
    'save bandwidth 9) start num bases. Sample type defaults: ' +
    ' || '.join((bst + ' : ' + ' '.join(map(str, params)))
                for bst, params in ALGN_PARAMS_TABLE.items())})
hidsigapars_opt=('--signal-align-parameters', {
    'type':float, 'nargs':len(next(iter(ALGN_PARAMS_TABLE.values()))),
    'help':argparse.SUPPRESS})


##############################
###### Choice arguments ######
##############################

normtype_opt=('--normalization-type', {
    'type':unicode,
    'default':'median', 'choices':('median', 'pA', 'pA_raw', 'none'),
    'help':'Choices: "none": raw 16-bit DAQ values, "pA_raw": pA as in the ' +
    'ONT events (using offset, range and digitization), "pA": k-mer-based ' +
    'correction for pA drift as in nanopolish (requires ' +
    '[--pore-model-filename]), "median": median and MAD from raw signal. ' +
    'Default: %(default)s'})

upstrmbs_opt=('--upstream-bases', {
    'default':1, 'type':int, 'choices':(0,1,2,3,4),
    'help':'Upstream bases in k-mer. Default: %(default)d'})
dnstrmbs_opt=('--downstream-bases', {
    'default':2, 'type':int, 'choices':(0,1,2,3,4),
    'help':'Downstream bases in k-mer. Default: %(default)d'})
altbs_opt=('--alternate-model-base', {
    'type':unicode, 'choices':('A','C','G','T'),
    'help':'Non-standard base is an alternative to this base.'})
modbs_opt=('--alternate-bases', {
    'type':unicode, 'choices':ALT_BASES, 'nargs':'+',
    'help':'Default non-standard base model for testing (not required if ' +
    'user created --alternate-model-filenames is provided).'})
paltmod_opt=('--plot-alternate-model', {
    'type':unicode, 'choices':ALT_BASES,
    'help':'Add alternative model distribution to the plot.'})
ovplttype_opt=('--overplot-type', {
    'type':unicode, 'default':'Downsample',
    'choices':['Downsample', 'Boxplot', 'Quantile', 'Density'],
    'help':'Plot type for regions with higher coverage. Default: Downsample'})
ftypes_opt=('--file-types', {
    'type':unicode, 'default':['coverage', ], 'nargs':'+',
    'choices':['coverage', 'valid_coverage', 'fraction', 'dampened_fraction',
               'signal', 'signal_sd', 'dwell', 'difference', 'statistic'],
    'help':'Data types of genome browser files to produce. Produced coverage ' +
    'files are in bedGraph format, while all other file types  will be in ' +
    'wiggle format (https://genome.ucsc.edu/goldenpath/help/wiggle.html). ' +
    'Default: "coverage"'})
stype_opt=('--statistic-type', {
    'type':unicode, 'default':'ks', 'choices':['ks', 'u', 't'],
    'help':'Type of statistical test to apply. Default: "ks"'})

dna_opt=('--dna', {
    'dest':'seq_sample_type', 'action':'store_const', 'const':DNA_SAMP_TYPE,
    'help':'Explicitly select canonical DNA model. Default: Automatically ' +
    'determine from FAST5s'})
rna_opt=('--rna', {
    'dest':'seq_sample_type', 'action':'store_const', 'const':RNA_SAMP_TYPE,
    'help':'Explicitly select canonical RNA model. Default: Automatically ' +
    'determine from FAST5s'})


###########################
###### Help argument ######
###########################

help_opt=(('--help', '-h'), {
    'action':'help',
    'help':"Print this help message and exit"})


##############################
###### Helper functions ######
##############################

OUTPUT_BASE = 'tombo_results'

def add_misc_args(parser):
    misc_args = parser.add_argument_group('Miscellaneous Arguments')
    misc_args.add_argument(*quiet_opt[0], **quiet_opt[1])
    misc_args.add_argument(*help_opt[0], **help_opt[1])

    return misc_args, parser

def add_common_testing_args(parser):
    io_args = parser.add_argument_group('Output Argument')
    io_args.add_argument(prstatbn_opt[0], **prstatbn_opt[1])
    io_args.add_argument(mstsgnf_opt[0], **mstsgnf_opt[1])

    multi_args = parser.add_argument_group('Multiprocessing Arguments')
    multi_args.add_argument(mpreg_opt[0], **mpreg_opt[1])
    multi_args.add_argument(proc_opt[0], default=1, **proc_opt[1])

    return io_args, multi_args

def add_default_args(parser):
    fast5_args = parser.add_argument_group('FAST5 Data Arguments')
    fast5_args.add_argument(corrgrp_opt[0], **corrgrp_opt[1])
    fast5_args.add_argument(bcsubgrps_opt[0], **bcsubgrps_opt[1])

    misc_args = add_misc_args(parser)

    return fast5_args, misc_args, parser

def add_comp_dist_args(parser):
    alt_args = parser.add_argument_group('Comparison Arguments')
    alt_args.add_argument(ctrlfast5dir_opt[0], **ctrlfast5dir_opt[1])
    alt_args.add_argument(pstdmod_opt[0], **pstdmod_opt[1])
    alt_args.add_argument(paltmod_opt[0], **paltmod_opt[1])
    alt_args.add_argument(hidden_tbmod_opt[0], **hidden_tbmod_opt[1])
    alt_args.add_argument(hidden_atbmod_opt[0], **hidden_atbmod_opt[1])

    return alt_args, parser


#####################################
###### Main re-squiggle parser ######
#####################################

def get_resquiggle_parser():
    parser = argparse.ArgumentParser(
        description='Re-segment raw nanopore signal to match with mapped ' +
        'portion of a known genomic sequence guided by a k-mer model.',
        add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(basedir_opt[0], **basedir_opt[1])
    req_args.add_argument(fasta_pos_opt[0], **fasta_pos_opt[1])

    mod_args = parser.add_argument_group('Model Parameters')
    mod_args.add_argument(dna_opt[0], **dna_opt[1])
    mod_args.add_argument(rna_opt[0], **rna_opt[1])

    filt_args = parser.add_argument_group('Read Filtering Argument')
    filt_args.add_argument(obsfilt_opt[0], **obsfilt_opt[1])
    filt_args.add_argument(qscr_opt[0], default=0, **qscr_opt[1])
    filt_args.add_argument(sms_opt[0], **sms_opt[1])

    multi_args = parser.add_argument_group('Multiprocessing Arguments')
    multi_args.add_argument(proc_opt[0], default=1, **proc_opt[1])
    multi_args.add_argument(hidthrpp_opt[0], **hidthrpp_opt[1])

    fast5_args = parser.add_argument_group('FAST5 Data Arguments')
    fast5_args.add_argument(corrgrp_opt[0], **corrgrp_opt[1])
    fast5_args.add_argument(bcgrp_opt[0], **bcgrp_opt[1])
    fast5_args.add_argument(bcsubgrps_opt[0], **bcsubgrps_opt[1])
    fast5_args.add_argument(ovrwrt_opt[0], **ovrwrt_opt[1])

    io_args = parser.add_argument_group('Input/Output Arguments')
    io_args.add_argument(failed_opt[0], **failed_opt[1])
    io_args.add_argument(dynerr_opt[0], **dynerr_opt[1])

    hid_args = parser.add_argument_group('Advanced Arguments')
    hid_args.add_argument(printadv_opt[0], **printadv_opt[1])
    hid_args.add_argument(hidden_tbmod_opt[0], **hidden_tbmod_opt[1])
    hid_args.add_argument(hidsegpars_opt[0], **hidsegpars_opt[1])
    hid_args.add_argument(hidsigapars_opt[0], **hidsigapars_opt[1])
    hid_args.add_argument(hidsss_opt[0], **hidsss_opt[1])
    hid_args.add_argument(hidmsi_opt[0], **hidmsi_opt[1])
    hid_args.add_argument(hidsigrng_opt[0], **hidsigrng_opt[1])
    hid_args.add_argument(hidseqrng_opt[0], **hidseqrng_opt[1])
    hid_args.add_argument(hidfitscl_opt[0], **hidfitscl_opt[1])
    hid_args.add_argument(hidfxdscl_opt[0], **hidfxdscl_opt[1])
    hid_args.add_argument(hidotlthresh_opt[0], **hidotlthresh_opt[1])
    hid_args.add_argument(hidskpidx_opt[0], **hidskpidx_opt[1])
    hid_args.add_argument(hidincldsd_opt[0], **hidincldsd_opt[1])
    hid_args.add_argument(hidignrlock_opt[0], **hidignrlock_opt[1])

    misc_args, parser = add_misc_args(parser)

    return parser

def print_advanced_resquiggle():
    parser = argparse.ArgumentParser(
        description='Hidden parameters to the resquiggle command.',
        add_help=False, usage='')
    hid_args = parser.add_argument_group('Hidden Arguments')
    hid_args.add_argument(tbmod_opt[0], **tbmod_opt[1])
    hid_args.add_argument(segpars_opt[0], **segpars_opt[1])
    hid_args.add_argument(sigapars_opt[0], **sigapars_opt[1])
    hid_args.add_argument(sss_opt[0], **sss_opt[1])
    hid_args.add_argument(msi_opt[0], **msi_opt[1])
    hid_args.add_argument(sigrng_opt[0], **sigrng_opt[1])
    hid_args.add_argument(seqrng_opt[0], **seqrng_opt[1])
    hid_args.add_argument(fitscl_opt[0], **fitscl_opt[1])
    hid_args.add_argument(fxdscl_opt[0], **fxdscl_opt[1])
    hid_args.add_argument(otlthresh_opt[0], **otlthresh_opt[1])
    hid_args.add_argument(skpidx_opt[0], **skpidx_opt[1])
    hid_args.add_argument(incldsd_opt[0], **incldsd_opt[1])
    hid_args.add_argument(ignrlock_opt[0], **ignrlock_opt[1])
    hid_args.add_argument(thrpp_opt[0], **thrpp_opt[1])

    hid_args.add_argument(*help_opt[0], **help_opt[1])

    return parser.parse_args(['-h',])


#############################################
###### Alternaitve re-squiggle parsers ######
#############################################

def get_event_resquiggle_parser():
    parser = argparse.ArgumentParser(
        description='Re-segment raw nanopore signal to match with mapped ' +
        'portion of a known genomic sequence guided by stored events ' +
        'from basecaller.', add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(basedir_opt[0], **basedir_opt[1])
    req_args.add_argument(fasta_event_opt[0], **fasta_event_opt[1])

    mapper_args = parser.add_argument_group(
        'Mapper Arguments (One mapper is required)')
    mapper_args.add_argument(minimap2_opt[0], **minimap2_opt[1])
    mapper_args.add_argument(minindx_opt[0], **minindx_opt[1])
    mapper_args.add_argument(bwamem_opt[0], **bwamem_opt[1])
    mapper_args.add_argument(graphmap_opt[0], **graphmap_opt[1])
    mapper_args.add_argument(batchsize_opt[0], **batchsize_opt[1])

    norm_args = parser.add_argument_group('Signal Processing Arguments')
    norm_args.add_argument(normtype_opt[0], **normtype_opt[1])
    norm_args.add_argument(poremod_opt[0], **poremod_opt[1])
    norm_args.add_argument(otlthresh_opt[0], **otlthresh_opt[1])
    norm_args.add_argument(segpars2_opt[0], **segpars2_opt[1])

    filt_args = parser.add_argument_group('Read Filtering Arguments')
    filt_args.add_argument(obsfilt_opt[0], **obsfilt_opt[1])
    filt_args.add_argument(timeout_opt[0], **timeout_opt[1])
    filt_args.add_argument(cpt_opt[0], **cpt_opt[1])

    io_args = parser.add_argument_group('Input/Output Arguments')
    io_args.add_argument(skpidx_opt[0], **skpidx_opt[1])
    io_args.add_argument(ovrwrt_opt[0], **ovrwrt_opt[1])
    io_args.add_argument(failed_opt[0], **failed_opt[1])
    io_args.add_argument(incldsd_opt[0], **incldsd_opt[1])

    fast5_args = parser.add_argument_group('FAST5 Data Arguments')
    fast5_args.add_argument(correvntgrp_opt[0], **correvntgrp_opt[1])
    fast5_args.add_argument(bcgrp_opt[0], **bcgrp_opt[1])
    fast5_args.add_argument(bcsubgrps_opt[0], **bcsubgrps_opt[1])

    multi_args = parser.add_argument_group('Multiprocessing Arguments')
    multi_args.add_argument(proc_opt[0], default=2, **proc_opt[1])
    multi_args.add_argument(alignproc_opt[0], **alignproc_opt[1])
    multi_args.add_argument(alignthrds_opt[0], **alignthrds_opt[1])
    multi_args.add_argument(rsqglproc_opt[0], **rsqglproc_opt[1])

    misc_args, parser = add_misc_args(parser)

    return parser


###################################
###### Pre-processing parser ######
###################################

def get_add_fastqs_parser():
    parser = argparse.ArgumentParser(
        description='Annotate raw FAST5 files with basecalls from a FASTQ.',
        add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(sfast5dir_opt[0], required=True, **sfast5dir_opt[1])
    req_args.add_argument(fastqs_opt[0], required=True, **fastqs_opt[1])

    fast5_args = parser.add_argument_group('FAST5 Data Arguments')
    fast5_args.add_argument(bcgrp_opt[0], **bcgrp_opt[1])
    fast5_args.add_argument(bcsubgrp_opt[0], **bcsubgrp_opt[1])
    fast5_args.add_argument(ovrwrt_opt[0], **ovrwrt_opt[1])

    seqsum_args = parser.add_argument_group('Sequencing Summary Argument')
    seqsum_args.add_argument(seqsum_opt[0], **seqsum_opt[1])

    multi_args = parser.add_argument_group('Multiprocessing Argument')
    multi_args.add_argument(proc_opt[0], default=1, **proc_opt[1])

    misc_args, parser = add_misc_args(parser)

    return parser


######################################
###### Model estimation parsers ######
######################################

def get_est_ref_parser():
    parser = argparse.ArgumentParser(
        description='Estimate standard tombo model for use in re-squiggle ' +
        'and  testing without an amplified (un-modified) sample.',
        add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])
    req_args.add_argument(tbmod_w_opt[0], required=True, **tbmod_w_opt[1])

    stat_args = parser.add_argument_group('Modeling Arguments')
    stat_args.add_argument(estmean_opt[0], **estmean_opt[1])
    stat_args.add_argument(kmspec_opt[0], **kmspec_opt[1])
    stat_args.add_argument(upstrmbs_opt[0], **upstrmbs_opt[1])
    stat_args.add_argument(dnstrmbs_opt[0], **dnstrmbs_opt[1])

    filt_args = parser.add_argument_group('Filtering Arguments')
    filt_args.add_argument(minreads_opt[0], default=10, **minreads_opt[1])
    filt_args.add_argument(covthresh_opt[0],  default=100, **covthresh_opt[1])
    filt_args.add_argument(minkmer_opt[0], default=5, **minkmer_opt[1])

    multi_args = parser.add_argument_group('Multiprocessing Arguments')
    multi_args.add_argument(mpreg_opt[0], **mpreg_opt[1])
    multi_args.add_argument(proc_opt[0], default=1, **proc_opt[1])

    fast5_args, misc_args, parser = add_default_args(parser)

    return parser

def get_est_alt_ref_parser():
    parser = argparse.ArgumentParser(
        description='Estimate alternate-base k-mer reference model for use ' +
        'in testing for specific modification types. [--fast5-basedirs] ' +
        'should contain a sample spiked with a single known randomly ' +
        'incorporated base.', add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(atbmod_opt[0], required=True, **atbmod_opt[1])
    req_args.add_argument(altname_opt[0], required=True, **altname_opt[1])
    req_args.add_argument(altbs_opt[0], required=True, **altbs_opt[1])

    dens_args = parser.add_argument_group(
        'Signal Data Arguments (Must provide either FAST5 dirs or ' +
        'previous density estimates)')
    dens_args.add_argument(fast5dir_opt[0], **fast5dir_opt[1])
    dens_args.add_argument(ctrlfast5dir_opt[0], **ctrlfast5dir_opt[1])
    dens_args.add_argument(altden_opt[0], **altden_opt[1])
    dens_args.add_argument(ctrlden_opt[0], **ctrlden_opt[1])

    mod_args = parser.add_argument_group('Standard Model Arguments')
    mod_args.add_argument(dna_opt[0], **dna_opt[1])
    mod_args.add_argument(rna_opt[0], **rna_opt[1])
    mod_args.add_argument(tbmod_opt[0], **tbmod_opt[1])

    stat_args = parser.add_argument_group('Model Fitting Arguments')
    stat_args.add_argument(altfrac_opt[0], **altfrac_opt[1])
    stat_args.add_argument(kernden_opt[0], **kernden_opt[1])

    filt_args = parser.add_argument_group('Filtering Argument')
    filt_args.add_argument(minkmer_opt[0], default=1000, **minkmer_opt[1])

    io_args = parser.add_argument_group('Output Argument')
    io_args.add_argument(densbn_opt[0], **densbn_opt[1])

    multi_args = parser.add_argument_group('Multiprocessing Arguments')
    multi_args.add_argument(proc_opt[0], default=1, **proc_opt[1])

    fast5_args, misc_args, parser = add_default_args(parser)

    return parser

def get_est_motif_alt_ref_parser():
    parser = argparse.ArgumentParser(
        description='Estimate motif-specific alternate-base k-mer reference ' +
        'model. [--fast5-basedirs] should contain a sample with a specific ' +
        'alternate-base occuring at a specific motif and ideally only ' +
        'canonical bases otherwise. These models often perform better than ' +
        'models trained from samples with random alternate-base incorporation.',
        add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(atbmod_opt[0], required=True, **atbmod_opt[1])
    req_args.add_argument(altname_opt[0], required=True, **altname_opt[1])
    req_args.add_argument(motifdesc_opt[0], required=True, **motifdesc_opt[1])
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])

    mod_args = parser.add_argument_group(
        'Modeling Arguments (Should match canoncial model)')
    mod_args.add_argument(upstrmbs_opt[0], **upstrmbs_opt[1])
    mod_args.add_argument(dnstrmbs_opt[0], **dnstrmbs_opt[1])

    filt_args = parser.add_argument_group('Filtering Argument')
    filt_args.add_argument(minkmer_opt[0], default=1, **minkmer_opt[1])
    filt_args.add_argument(validlocs_opt[0], **validlocs_opt[1])
    filt_args.add_argument(minreads_opt[0], default=10, **minreads_opt[1])
    filt_args.add_argument(covthresh_opt[0],  default=100, **covthresh_opt[1])

    multi_args = parser.add_argument_group('Multiprocessing Arguments')
    multi_args.add_argument(mpreg_opt[0], **mpreg_opt[1])
    multi_args.add_argument(proc_opt[0], default=1, **proc_opt[1])

    fast5_args, misc_args, parser = add_default_args(parser)

    return parser

def get_estimate_scale_parser():
    parser = argparse.ArgumentParser(
        description='Estimate a global scaling parameter from a ' +
        'sub-set of reads.', add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(basedir_opt[0], **basedir_opt[1])

    misc_args = parser.add_argument_group('Miscellaneous Arguments')
    misc_args.add_argument(*quiet_opt[0], **quiet_opt[1])
    misc_args.add_argument(*help_opt[0], **help_opt[1])

    return parser


#########################################
###### Significance testing parser ######
#########################################

def get_de_novo_test_signif_parser():
    parser = argparse.ArgumentParser(
        description='Test for significant shifts in raw nanopore signal ' +
        'awway from a canonical base expected signal model.', add_help=False)
    req_args = parser.add_argument_group('Required Argument')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])
    req_args.add_argument(statbsnm_opt[0], required=True, **statbsnm_opt[1])

    alt_args = parser.add_argument_group('Comparison Model Arguments')
    alt_args.add_argument(dna_opt[0], **dna_opt[1])
    alt_args.add_argument(rna_opt[0], **rna_opt[1])
    alt_args.add_argument(hidden_tbmod_opt[0], **hidden_tbmod_opt[1])

    test_args = parser.add_argument_group('Significance Test Arguments')
    test_args.add_argument(fmo_opt[0], **fmo_opt[1])
    test_args.add_argument(minreads_opt[0], default=1, **minreads_opt[1])
    test_args.add_argument(dnthresh_opt[0], **dnthresh_opt[1])
    test_args.add_argument(cvgdmp_opt[0], **cvgdmp_opt[1])

    io_args, multi_args = add_common_testing_args(parser)
    fast5_args, misc_args, parser = add_default_args(parser)

    return parser

def get_alt_test_signif_parser():
    parser = argparse.ArgumentParser(
        description='Test for significant shifts in raw nanopore signal ' +
        'specificly matching a non-canonical base model.', add_help=False)
    req_args = parser.add_argument_group('Required Argument')
    req_args.add_argument(fast5dir_opt[0], **fast5dir_opt[1])
    req_args.add_argument(statbsnm_opt[0], **statbsnm_opt[1])
    req_args.add_argument(modbs_opt[0], **modbs_opt[1])

    alt_args = parser.add_argument_group('Comparison Arguments')
    alt_args.add_argument(printalt_opt[0], **printalt_opt[1])
    alt_args.add_argument(dna_opt[0], **dna_opt[1])
    alt_args.add_argument(rna_opt[0], **rna_opt[1])
    alt_args.add_argument(hidden_tbmod_opt[0], **hidden_tbmod_opt[1])
    alt_args.add_argument(hidden_atbmods_opt[0], **hidden_atbmods_opt[1])

    test_args = parser.add_argument_group('Significance Test Arguments')
    test_args.add_argument(minreads_opt[0], default=1, **minreads_opt[1])
    test_args.add_argument(altthresh_opt[0], **altthresh_opt[1])
    test_args.add_argument(stdllhr_opt[0], **stdllhr_opt[1])
    test_args.add_argument(cvgdmp_opt[0], **cvgdmp_opt[1])

    io_args, multi_args = add_common_testing_args(parser)
    fast5_args, misc_args, parser = add_default_args(parser)

    return parser

def get_samp_comp_test_signif_parser():
    parser = argparse.ArgumentParser(
        description='Test for significant shifts in raw nanopore signal ' +
        'from levels estimated from a control/canonical base only sample ' +
        '(usually PCR for DNA or IVT for RNA).', add_help=False)
    req_args = parser.add_argument_group('Required Argument')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])
    req_args.add_argument(statbsnm_opt[0], required=True, **statbsnm_opt[1])
    req_args.add_argument(
        ctrlfast5dir_opt[0], required=True, **ctrlfast5dir_opt[1])

    alt_args = parser.add_argument_group('Model Prior Arguments')
    alt_args.add_argument(samponly_opt[0], **samponly_opt[1])
    alt_args.add_argument(prwht_opt[0], **prwht_opt[1])
    alt_args.add_argument(dna_opt[0], **dna_opt[1])
    alt_args.add_argument(rna_opt[0], **rna_opt[1])
    alt_args.add_argument(hidden_tbmod_opt[0], **hidden_tbmod_opt[1])

    test_args = parser.add_argument_group('Significance Test Arguments')
    test_args.add_argument(fmo_opt[0], **fmo_opt[1])
    test_args.add_argument(minreads_opt[0], default=1, **minreads_opt[1])
    test_args.add_argument(scompthresh_opt[0], **scompthresh_opt[1])
    test_args.add_argument(cvgdmp_opt[0], **cvgdmp_opt[1])

    io_args, multi_args = add_common_testing_args(parser)
    fast5_args, misc_args, parser = add_default_args(parser)

    return parser

def get_group_comp_test_signif_parser():
    parser = argparse.ArgumentParser(
        description='Test for significant shifts in raw nanopore signal ' +
        'away from a control/canonical base only sample (usually ' +
        'PCR for DNA or IVT for RNA).', add_help=False)
    req_args = parser.add_argument_group('Required Argument')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])
    req_args.add_argument(statbsnm_opt[0], required=True, **statbsnm_opt[1])
    req_args.add_argument(
        altfast5dir_opt[0], required=True, **altfast5dir_opt[1])

    test_args = parser.add_argument_group('Significance Test Arguments')
    test_args.add_argument(fmo_opt[0], **fmo_opt[1])
    test_args.add_argument(minreads_opt[0], default=50, **minreads_opt[1])
    test_args.add_argument(stype_opt[0], **stype_opt[1])
    test_args.add_argument(storepval_opt[0], **storepval_opt[1])

    io_args = parser.add_argument_group('Output Argument')
    io_args.add_argument(mstsgnf_opt[0], **mstsgnf_opt[1])

    multi_args = parser.add_argument_group('Multiprocessing Arguments')
    multi_args.add_argument(mpreg_opt[0], **mpreg_opt[1])
    multi_args.add_argument(proc_opt[0], default=1, **proc_opt[1])

    fast5_args, misc_args, parser = add_default_args(parser)

    return parser

def get_aggregate_per_read_parser():
    parser = argparse.ArgumentParser(
        description='Aggregate per-read statistics to produce a standard ' +
        '(genomic base) statistics file.', add_help=False)
    req_args = parser.add_argument_group('Required Argument')
    req_args.add_argument(prstat_opt[0], required=True, **prstat_opt[1])
    req_args.add_argument(statfn_opt[0], required=True, **statfn_opt[1])
    req_args.add_argument(snglrdthrsh_opt[0], required=True, **snglrdthrsh_opt[1])

    test_args = parser.add_argument_group('Significance Test Arguments')
    test_args.add_argument(minreads_opt[0], default=1, **minreads_opt[1])
    test_args.add_argument(cvgdmp_opt[0], **cvgdmp_opt[1])

    io_args = parser.add_argument_group('Output Argument')
    io_args.add_argument(mstsgnf_opt[0], **mstsgnf_opt[1])

    multi_args = parser.add_argument_group('Multiprocessing Arguments')
    multi_args.add_argument(proc_opt[0], default=1, **proc_opt[1])

    fast5_args, misc_args, parser = add_default_args(parser)

    return parser


############################
###### Filter parsers ######
############################

def get_clear_filters_parser():
    parser = argparse.ArgumentParser(
        description='Clear all filters applied to re-squiggled reads.',
        add_help=False)
    req_args = parser.add_argument_group('Required Argument')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])

    fast5_args = parser.add_argument_group('FAST5 Data Argument')
    fast5_args.add_argument(corrgrp_opt[0], **corrgrp_opt[1])

    misc_args, parser = add_misc_args(parser)

    return parser

def get_filter_stuck_parser():
    parser = argparse.ArgumentParser(
        description='Filter reads based on observations ' +
        'per base thresholds.', add_help=False)
    req_args = parser.add_argument_group('Required Argument')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])

    filter_args = parser.add_argument_group('Read Filtering Argument')
    filter_args.add_argument(obsfilt_opt[0], **obsfilt_opt[1])

    fast5_args = parser.add_argument_group('FAST5 Data Argument')
    fast5_args.add_argument(corrgrp_opt[0], **corrgrp_opt[1])

    misc_args, parser = add_misc_args(parser)

    return parser

def get_filter_coverage_parser():
    parser = argparse.ArgumentParser(
        description='Filter reads by downsampling for more even coverage.',
        add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])

    filter_args = parser.add_argument_group('Read Filtering Argument')
    filter_args.add_argument(pctfilt_opt[0], **pctfilt_opt[1])

    fast5_args = parser.add_argument_group('FAST5 Data Arguments')
    fast5_args.add_argument(corrgrp_opt[0], **corrgrp_opt[1])

    misc_args, parser = add_misc_args(parser)

    return parser

def get_filter_qscore_parser():
    parser = argparse.ArgumentParser(
        description='Filter reads to remove low quality reads.',
        add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])

    filter_args = parser.add_argument_group('Read Filtering Argument')
    filter_args.add_argument(qscr_opt[0], default=7, **qscr_opt[1])

    fast5_args = parser.add_argument_group('FAST5 Data Arguments')
    fast5_args.add_argument(corrgrp_opt[0], **corrgrp_opt[1])
    fast5_args.add_argument(bcgrp_opt[0], **bcgrp_opt[1])

    misc_args, parser = add_misc_args(parser)

    return parser

def get_filter_signal_matching_parser():
    parser = argparse.ArgumentParser(
        description='Filter reads with poor raw to expected signal matching.',
        add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])
    req_args.add_argument(sms_opt[0], required=True, **sms_opt[1])

    fast5_args = parser.add_argument_group('FAST5 Data Arguments')
    fast5_args.add_argument(corrgrp_opt[0], **corrgrp_opt[1])

    misc_args, parser = add_misc_args(parser)

    return parser

def get_filter_genome_pos_parser():
    parser = argparse.ArgumentParser(
        description='Filter reads based on genome mapping location.',
        add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])
    req_args.add_argument(incldreg_opt[0], **incldreg_opt[1])

    filt_args = parser.add_argument_group('Filter Argument')
    filt_args.add_argument(prtovlp_opt[0], **prtovlp_opt[1])

    fast5_args = parser.add_argument_group('FAST5 Data Argument')
    fast5_args.add_argument(corrgrp_opt[0], **corrgrp_opt[1])

    misc_args, parser = add_misc_args(parser)

    return parser


##############################################
###### Genome-anchored plotting parsers ######
##############################################

def get_max_cov_parser():
    parser = argparse.ArgumentParser(
        description='Plot raw signal in regions with the deepest coverage.',
        add_help=False)
    req_args = parser.add_argument_group('Required Argument')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])

    alt_args, parser = add_comp_dist_args(parser)

    ovplt_args = parser.add_argument_group('Overplotting Arguments')
    ovplt_args.add_argument(ovpltthresh_opt[0], **ovpltthresh_opt[1])
    ovplt_args.add_argument(ovplttype_opt[0], **ovplttype_opt[1])

    reg_args = parser.add_argument_group('Plotting Region Arguments')
    reg_args.add_argument(numreg_opt[0], default=10, **numreg_opt[1])
    reg_args.add_argument(numbases_opt[0], default=21, **numbases_opt[1])

    out_args = parser.add_argument_group('Output Argument')
    out_args.add_argument(pdf_opt[0], default=OUTPUT_BASE + '.max_coverage.pdf',
                          **pdf_opt[1])

    fast5_args, misc_args, parser = add_default_args(parser)

    return parser

def get_genome_loc_parser():
    parser = argparse.ArgumentParser(
        description='Plot raw signal at defined genomic locations.',
        add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])
    req_args.add_argument(gnmloc_opt[0], required=True, **gnmloc_opt[1])

    alt_args, parser = add_comp_dist_args(parser)

    ovplt_args = parser.add_argument_group('Overplotting Arguments')
    ovplt_args.add_argument(ovpltthresh_opt[0], **ovpltthresh_opt[1])
    ovplt_args.add_argument(ovplttype_opt[0], **ovplttype_opt[1])

    reg_args = parser.add_argument_group('Plotting Region Argument')
    reg_args.add_argument(numbases_opt[0], default=21, **numbases_opt[1])

    out_args = parser.add_argument_group('Output Argument')
    out_args.add_argument(pdf_opt[0],
                          default=OUTPUT_BASE + '.genome_locations.pdf',
                          **pdf_opt[1])

    fast5_args, misc_args, parser = add_default_args(parser)

    return parser

def get_motif_loc_parser():
    parser = argparse.ArgumentParser(
        description='Plot raw signal centered on a motif of interest.',
        add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])
    req_args.add_argument(motif_opt[0], required=True, **motif_opt[1])
    req_args.add_argument(fasta_opt[0], required=True, **fasta_opt[1])

    alt_args, parser = add_comp_dist_args(parser)

    ovplt_args = parser.add_argument_group('Overplotting Arguments')
    ovplt_args.add_argument(ovpltthresh_opt[0], **ovpltthresh_opt[1])
    ovplt_args.add_argument(ovplttype_opt[0], **ovplttype_opt[1])

    reg_args = parser.add_argument_group('Plotting Region Arguments')
    reg_args.add_argument(numreg_opt[0], default=10, **numreg_opt[1])
    reg_args.add_argument(numbases_opt[0], default=21, **numbases_opt[1])
    reg_args.add_argument(deepcov_opt[0], **deepcov_opt[1])

    out_args = parser.add_argument_group('Output Argument')
    out_args.add_argument(pdf_opt[0],
                          default=OUTPUT_BASE + '.motif_centered.pdf',
                          **pdf_opt[1])

    fast5_args, misc_args, parser = add_default_args(parser)

    return parser

def get_max_diff_parser():
    parser = argparse.ArgumentParser(
        description='Plot raw signal where the average signal differs ' +
        'most between two samples.', add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])
    req_args.add_argument(ctrlfast5dir_opt[0], required=True,
                          **ctrlfast5dir_opt[1])

    ovplt_args = parser.add_argument_group('Overplotting Arguments')
    ovplt_args.add_argument(ovpltthresh_opt[0], **ovpltthresh_opt[1])
    ovplt_args.add_argument(ovplttype_opt[0], **ovplttype_opt[1])

    reg_args = parser.add_argument_group('Plotting Region Arguments')
    reg_args.add_argument(numreg_opt[0], default=10, **numreg_opt[1])
    reg_args.add_argument(numbases_opt[0], default=21, **numbases_opt[1])

    out_args = parser.add_argument_group('Output Arguments')
    out_args.add_argument(pdf_opt[0], default=OUTPUT_BASE + '.max_difference.pdf',
                          **pdf_opt[1])
    out_args.add_argument(seqs_opt[0], **seqs_opt[1])

    fast5_args, misc_args, parser = add_default_args(parser)

    return parser

def get_signif_diff_parser():
    parser = argparse.ArgumentParser(
        description='Plot raw signal at most significant genomic locations ' +
        'from previous test_significance results.', add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])
    req_args.add_argument(statfn_opt[0], required=True, **statfn_opt[1])

    alt_args, parser = add_comp_dist_args(parser)

    ovplt_args = parser.add_argument_group('Overplotting Arguments')
    ovplt_args.add_argument(ovpltthresh_opt[0], **ovpltthresh_opt[1])
    ovplt_args.add_argument(ovplttype_opt[0], **ovplttype_opt[1])

    reg_args = parser.add_argument_group('Plotting Region Arguments')
    reg_args.add_argument(numreg_opt[0], default=10, **numreg_opt[1])
    reg_args.add_argument(numbases_opt[0], default=21, **numbases_opt[1])

    stat_args = parser.add_argument_group('Statistical Argument')
    stat_args.add_argument(cvgdmp_opt[0], **cvgdmp_opt[1])

    out_args = parser.add_argument_group('Output Arguments')
    out_args.add_argument(pdf_opt[0],
                          default=OUTPUT_BASE + '.significant_difference.pdf',
                          **pdf_opt[1])
    out_args.add_argument(seqs_opt[0], **seqs_opt[1])

    fast5_args, misc_args, parser = add_default_args(parser)

    return parser

def get_signif_motif_parser():
    parser = argparse.ArgumentParser(
        description='Plot raw signal centered on motif of interest along ' +
        'with test statistic distributions at relative genomic position.',
        add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])
    req_args.add_argument(motif_opt[0], required=True, **motif_opt[1])
    req_args.add_argument(statfn_opt[0], required=True, **statfn_opt[1])
    req_args.add_argument(fasta_opt[0], required=True,  **fasta_opt[1])

    alt_args, parser = add_comp_dist_args(parser)

    ovplt_args = parser.add_argument_group('Overplotting Argument')
    ovplt_args.add_argument(ovpltthresh_opt[0], **ovpltthresh_opt[1])

    reg_args = parser.add_argument_group('Plotting Region Arguments')
    reg_args.add_argument(numreg_opt[0], default=3, **numreg_opt[1])
    reg_args.add_argument(cntxt_opt[0], **cntxt_opt[1])
    reg_args.add_argument(numstat_opt[0], **numstat_opt[1])

    stat_args = parser.add_argument_group('Statistical Argument')
    stat_args.add_argument(cvgdmp_opt[0], **cvgdmp_opt[1])

    out_args = parser.add_argument_group('Output Argument')
    out_args.add_argument(pdf_opt[0],
                          default=OUTPUT_BASE + '.motif_statistics.pdf',
                          **pdf_opt[1])

    fast5_args, misc_args, parser = add_default_args(parser)

    return parser

def get_per_read_parser():
    parser = argparse.ArgumentParser(
        description='Plot non-standard base statistic per read at specified ' +
        'genomic locations.', add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(gnmloc_opt[0], required=True, **gnmloc_opt[1])
    req_args.add_argument(prstat_opt[0], required=True, **prstat_opt[1])

    seq_args = parser.add_argument_group(
        'Sequence Arguments (Provide either FAST5s dir or genome FASTA)')
    seq_args.add_argument(fasta_opt[0], **fasta_opt[1])
    seq_args.add_argument(fast5dir_opt[0], **fast5dir_opt[1])

    reg_args = parser.add_argument_group('Plotting Region Arguments')
    reg_args.add_argument(numreads_opt[0], default=100, **numreads_opt[1])
    reg_args.add_argument(numbases_opt[0], default=51, **numbases_opt[1])
    reg_args.add_argument(boxc_opt[0], **boxc_opt[1])

    out_args = parser.add_argument_group('Output Argument')
    out_args.add_argument(
        pdf_opt[0], default=OUTPUT_BASE + '.per_read_stats.pdf', **pdf_opt[1])

    fast5_args, misc_args, parser = add_default_args(parser)

    return parser


####################################
###### Other plotting parsers ######
####################################

def get_kmer_dist_parser():
    parser = argparse.ArgumentParser(
        description='Plot signal level distribution across kmers.',
        add_help=False)
    req_args = parser.add_argument_group('Required Argument')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])

    proc_args = parser.add_argument_group('Data Processing Arguments')
    proc_args.add_argument(upstrmbs_opt[0], **upstrmbs_opt[1])
    proc_args.add_argument(dnstrmbs_opt[0], **dnstrmbs_opt[1])
    proc_args.add_argument(readmean_opt[0], **readmean_opt[1])
    proc_args.add_argument(kmerthresh_opt[0], **kmerthresh_opt[1])

    reg_args = parser.add_argument_group('Plotting Region Arguments')
    reg_args.add_argument(numreads_opt[0], default=100, **numreads_opt[1])

    out_args = parser.add_argument_group('Output Arguments')
    out_args.add_argument(pdf_opt[0],
                         default=OUTPUT_BASE + '.kmer_distribution.pdf',
                         **pdf_opt[1])
    out_args.add_argument(rdata_opt[0], **rdata_opt[1])
    out_args.add_argument(noplot_opt[0], **noplot_opt[1])

    fast5_args, misc_args, parser = add_default_args(parser)

    return parser

def get_roc_parser():
    parser = argparse.ArgumentParser(
        description='Plot ROC curve given known motif(s).',
        add_help=False)
    req_args = parser.add_argument_group('Required Argument')
    req_args.add_argument(statfns_opt[0], required=True, **statfns_opt[1])

    gt_args = parser.add_argument_group(
        'Ground Truth Arguments (provide bed files or motifs)')
    gt_args.add_argument(moddescs_opt[0], **moddescs_opt[1])
    gt_args.add_argument(unmodlocs_opt[0], **unmodlocs_opt[1])
    gt_args.add_argument(motifdescs_opt[0], **motifdescs_opt[1])
    gt_args.add_argument(fasta_opt[0], **fasta_opt[1])

    out_args = parser.add_argument_group('Output Arguments')
    out_args.add_argument(
        pdf_opt[0], default=OUTPUT_BASE + '.roc.pdf', **pdf_opt[1])

    limit_args = parser.add_argument_group('Down-sampling Arguments')
    limit_args.add_argument(allspb_opt[0], **allspb_opt[1])
    limit_args.add_argument(tsl_opt[0], default=5000000, **tsl_opt[1])

    misc_args, parser = add_misc_args(parser)

    return parser

def get_control_roc_parser():
    parser = argparse.ArgumentParser(
        description='Plot ROC curve comparing a control and native sample ' +
        'at known motif(s).', add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(statfns_opt[0], required=True, **statfns_opt[1])
    req_args.add_argument(
        ctrlstatfns_opt[0], required=True, **ctrlstatfns_opt[1])
    req_args.add_argument(motifdescs_opt[0], required=True, **motifdescs_opt[1])
    req_args.add_argument(fasta_opt[0], required=True, **fasta_opt[1])

    out_args = parser.add_argument_group('Output Arguments')
    out_args.add_argument(
        pdf_opt[0], default=OUTPUT_BASE + '.sample_compare.roc.pdf',
        **pdf_opt[1])

    limit_args = parser.add_argument_group('Down-sampling Arguments')
    limit_args.add_argument(allspb_opt[0], **allspb_opt[1])
    limit_args.add_argument(tsl_opt[0], default=5000000, **tsl_opt[1])

    misc_args, parser = add_misc_args(parser)

    return parser

def get_per_read_roc_parser():
    parser = argparse.ArgumentParser(
        description='Plot per-read ROC curve given known motif(s).',
        add_help=False)
    req_args = parser.add_argument_group('Required Argument')
    req_args.add_argument(prstats_opt[0], required=True, **prstats_opt[1])

    gt_args = parser.add_argument_group(
        'Ground Truth Arguments (provide bed files or motifs)')
    gt_args.add_argument(moddescs_opt[0], **moddescs_opt[1])
    gt_args.add_argument(unmodlocs_opt[0], **unmodlocs_opt[1])
    gt_args.add_argument(motifdescs_opt[0], **motifdescs_opt[1])
    gt_args.add_argument(fasta_opt[0], **fasta_opt[1])

    limit_args = parser.add_argument_group('Down-sampling Arguments')
    limit_args.add_argument(spb_opt[0], default=100000, **spb_opt[1])
    limit_args.add_argument(tsl_opt[0], default=5000000, **tsl_opt[1])

    out_args = parser.add_argument_group('Output Arguments')
    out_args.add_argument(pdf_opt[0],
                         default=OUTPUT_BASE + '.per_reads_roc.pdf',
                         **pdf_opt[1])

    misc_args, parser = add_misc_args(parser)

    return parser

def get_control_per_read_roc_parser():
    parser = argparse.ArgumentParser(
        description='Plot per-read ROC curve comparing a control and ' +
        'native sample at known motif(s).', add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(prstats_opt[0], required=True, **prstats_opt[1])
    req_args.add_argument(
        prctrlstats_opt[0], required=True, **prctrlstats_opt[1])
    req_args.add_argument(motifdescs_opt[0], required=True, **motifdescs_opt[1])
    req_args.add_argument(fasta_opt[0], required=True, **fasta_opt[1])

    out_args = parser.add_argument_group('Output Arguments')
    out_args.add_argument(
        pdf_opt[0], default=OUTPUT_BASE + '.sample_compare.roc.pdf',
        **pdf_opt[1])

    limit_args = parser.add_argument_group('Down-sampling Arguments')
    limit_args.add_argument(allspb_opt[0], **allspb_opt[1])
    limit_args.add_argument(tsl_opt[0], default=5000000, **tsl_opt[1])

    misc_args, parser = add_misc_args(parser)

    return parser

def get_cluster_signif_diff_parser():
    parser = argparse.ArgumentParser(
        description='Cluster signal trace differences at most significant ' +
        'signal shifts from previous test_significance results.',
        add_help=False)
    req_args = parser.add_argument_group('Required Arguments')
    req_args.add_argument(fast5dir_opt[0], required=True, **fast5dir_opt[1])
    req_args.add_argument(ctrlfast5dir_opt[0], required=True,
                          **ctrlfast5dir_opt[1])
    req_args.add_argument(statfn_opt[0], required=True, **statfn_opt[1])

    fasta_args = parser.add_argument_group('FASTA Sequence Argument')
    fasta_args.add_argument(fasta_opt[0], **fasta_opt[1])

    multi_args = parser.add_argument_group('Multiprocessing Argument')
    multi_args.add_argument(proc_opt[0], default=1, **proc_opt[1])

    reg_args = parser.add_argument_group('Plotting Region Arguments')
    reg_args.add_argument(numreg_opt[0], default=10, **numreg_opt[1])
    reg_args.add_argument(numbases_opt[0], default=21, **numbases_opt[1])
    reg_args.add_argument(slides_opt[0], **slides_opt[1])

    out_args = parser.add_argument_group('Output Arguments')
    out_args.add_argument(pdf_opt[0],
                          default=OUTPUT_BASE + '.signal_clusters.pdf',
                          **pdf_opt[1])
    out_args.add_argument(rdata_opt[0], **rdata_opt[1])

    fast5_args, misc_args, parser = add_default_args(parser)

    return parser


#################################
###### Text output parsers ######
#################################

def get_browser_files_parser():
    parser = argparse.ArgumentParser(
        description='Write wiggle files for specified data types.',
        add_help=False)
    data_args = parser.add_argument_group('Data Arguments')
    data_args.add_argument(fast5dir_opt[0], **fast5dir_opt[1])
    data_args.add_argument(ctrlfast5dir_opt[0], **ctrlfast5dir_opt[1])
    data_args.add_argument(statfn_opt[0], **statfn_opt[1])

    motif_args = parser.add_argument_group('Statistic Motif Filter Arguments')
    motif_args.add_argument(fasta_opt[0], **fasta_opt[1])
    motif_args.add_argument(motifdescsimp_opt[0], **motifdescsimp_opt[1])

    out_args = parser.add_argument_group('Output Arguments')
    out_args.add_argument(brsrfn_opt[0], **brsrfn_opt[1])
    out_args.add_argument(ftypes_opt[0], **ftypes_opt[1])

    fast5_args, misc_args, parser = add_default_args(parser)

    return parser

def get_write_signif_diff_parser():
    parser = argparse.ArgumentParser(
        description='Write sequence at genomic locations with most ' +
        'significant difference from previous test_significance results.',
        add_help=False)
    req_args = parser.add_argument_group('Required Argument')
    req_args.add_argument(statfn_opt[0], required=True, **statfn_opt[1])

    seq_args = parser.add_argument_group(
        'Sequence Arguments (Provide either FAST5s dir or genome FASTA)')
    seq_args.add_argument(fasta_opt[0], **fasta_opt[1])
    seq_args.add_argument(fast5dir_opt[0], **fast5dir_opt[1])

    reg_args = parser.add_argument_group('Region Selection Arguments')
    reg_args.add_argument(numreg_opt[0], default=100, **numreg_opt[1])
    reg_args.add_argument(numbases_opt[0], default=15, **numbases_opt[1])

    out_args = parser.add_argument_group('Output Arguments')
    out_args.add_argument(
        seqs_opt[0], default=OUTPUT_BASE + '.significant_regions.fasta',
        **seqs_opt[1])

    fast5_args, misc_args, parser = add_default_args(parser)

    return parser


if __name__ == '__main__':
    sys.stderr.write('This is a module. See commands with `tombo -h`')
    sys.exit(1)
